import warnings
from types import MappingProxyType
from typing import (
    Any,
    Iterable,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
)

from tqdm import tqdm

from bioimageio.spec.model import AnyModelDescr, v0_4, v0_5
from bioimageio.spec.model.v0_5 import WeightsFormat

from ._op_base import BlockedOperator
from .axis import AxisId, PerAxis
from .common import Halo, MemberId, PerMember, SampleId
from .digest_spec import (
    get_block_transform,
    get_input_halo,
    get_member_ids,
)
from .model_adapters import ModelAdapter, create_model_adapter
from .model_adapters import get_weight_formats as get_weight_formats
from .proc_ops import Processing
from .proc_setup import setup_pre_and_postprocessing
from .sample import Sample, SampleBlock, SampleBlockWithOrigin
from .stat_measures import DatasetMeasure, MeasureValue, Stat
from .tensor import Tensor

Predict_IO = TypeVar(
    "Predict_IO",
    Sample,
    Iterable[Sample],
)


class PredictionPipeline:
    """
    Represents model computation including preprocessing and postprocessing
    Note: Ideally use the PredictionPipeline as a context manager
    """

    def __init__(
        self,
        *,
        name: str,
        model_description: AnyModelDescr,
        preprocessing: List[Processing],
        postprocessing: List[Processing],
        model_adapter: ModelAdapter,
        default_ns: Union[
            v0_5.ParameterizedSize_N,
            Mapping[Tuple[MemberId, AxisId], v0_5.ParameterizedSize_N],
        ] = 10,
        default_batch_size: int = 1,
    ) -> None:
        super().__init__()
        if model_description.run_mode:
            warnings.warn(
                f"Not yet implemented inference for run mode '{model_description.run_mode.name}'"
            )

        self.name = name
        self._preprocessing = preprocessing
        self._postprocessing = postprocessing

        self.model_description = model_description
        if isinstance(model_description, v0_4.ModelDescr):
            self._default_input_halo: PerMember[PerAxis[Halo]] = {}
            self._block_transform = None
        else:
            default_output_halo = {
                t.id: {
                    a.id: Halo(a.halo, a.halo)
                    for a in t.axes
                    if isinstance(a, v0_5.WithHalo)
                }
                for t in model_description.outputs
            }
            self._default_input_halo = get_input_halo(
                model_description, default_output_halo
            )
            self._block_transform = get_block_transform(model_description)

        self._default_ns = default_ns
        self._default_batch_size = default_batch_size

        self._input_ids = get_member_ids(model_description.inputs)
        self._output_ids = get_member_ids(model_description.outputs)

        self._adapter: ModelAdapter = model_adapter

    def __enter__(self):
        self.load()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):  # type: ignore
        self.unload()
        return False

    def predict_sample_block(
        self,
        sample_block: SampleBlockWithOrigin,
        skip_preprocessing: bool = False,
        skip_postprocessing: bool = False,
    ) -> SampleBlock:
        if isinstance(self.model_description, v0_4.ModelDescr):
            raise NotImplementedError(
                f"predict_sample_block not implemented for model {self.model_description.format_version}"
            )
        else:
            assert self._block_transform is not None

        if not skip_preprocessing:
            self.apply_preprocessing(sample_block)

        output_meta = sample_block.get_transformed_meta(self._block_transform)
        output = output_meta.with_data(
            {
                tid: out
                for tid, out in zip(
                    self._output_ids,
                    self._adapter.forward(
                        *(sample_block.members.get(t) for t in self._input_ids)
                    ),
                )
                if out is not None
            },
            stat=sample_block.stat,
        )
        if not skip_postprocessing:
            self.apply_postprocessing(output)

        return output

    def predict_sample_without_blocking(
        self,
        sample: Sample,
        skip_preprocessing: bool = False,
        skip_postprocessing: bool = False,
    ) -> Sample:
        """predict a sample.
        The sample's tensor shapes have to match the model's input tensor description.
        If that is not the case, consider `predict_sample_with_blocking`"""

        if not skip_preprocessing:
            self.apply_preprocessing(sample)

        output = Sample(
            members={
                out_id: out
                for out_id, out in zip(
                    self._output_ids,
                    self._adapter.forward(
                        *(sample.members.get(in_id) for in_id in self._input_ids)
                    ),
                )
                if out is not None
            },
            stat=sample.stat,
            id=self.get_output_sample_id(sample.id),
        )
        if not skip_postprocessing:
            self.apply_postprocessing(output)

        return output

    def get_output_sample_id(self, input_sample_id: SampleId):
        if input_sample_id is None:
            return None
        else:
            return f"{input_sample_id}_" + (
                self.model_description.id or self.model_description.name
            )

    def predict_sample_with_fixed_blocking(
        self,
        sample: Sample,
        input_block_shape: Mapping[MemberId, Mapping[AxisId, int]],
        *,
        skip_preprocessing: bool = False,
        skip_postprocessing: bool = False,
    ) -> Sample:
        if not skip_preprocessing:
            self.apply_preprocessing(sample)

        n_blocks, input_blocks = sample.split_into_blocks(
            input_block_shape,
            halo=self._default_input_halo,
            pad_mode="reflect",
        )
        input_blocks = list(input_blocks)
        predicted_blocks: List[SampleBlock] = []
        for b in tqdm(
            input_blocks,
            desc=f"predict sample {sample.id or ''} with {self.model_description.id or self.model_description.name}",
            unit="block",
            unit_divisor=1,
            total=n_blocks,
        ):
            predicted_blocks.append(
                self.predict_sample_block(
                    b, skip_preprocessing=True, skip_postprocessing=True
                )
            )

        predicted_sample = Sample.from_blocks(predicted_blocks)
        if not skip_postprocessing:
            self.apply_postprocessing(predicted_sample)

        return predicted_sample

    def predict_sample_with_blocking(
        self,
        sample: Sample,
        skip_preprocessing: bool = False,
        skip_postprocessing: bool = False,
        ns: Optional[
            Union[
                v0_5.ParameterizedSize_N,
                Mapping[Tuple[MemberId, AxisId], v0_5.ParameterizedSize_N],
            ]
        ] = None,
        batch_size: Optional[int] = None,
    ) -> Sample:
        """predict a sample by splitting it into blocks according to the model and the `ns` parameter"""

        if isinstance(self.model_description, v0_4.ModelDescr):
            raise NotImplementedError(
                "`predict_sample_with_blocking` not implemented for v0_4.ModelDescr"
                + f" {self.model_description.name}."
                + " Consider using `predict_sample_with_fixed_blocking`"
            )

        ns = ns or self._default_ns
        if isinstance(ns, int):
            ns = {
                (ipt.id, a.id): ns
                for ipt in self.model_description.inputs
                for a in ipt.axes
                if isinstance(a.size, v0_5.ParameterizedSize)
            }
        input_block_shape = self.model_description.get_tensor_sizes(
            ns, batch_size or self._default_batch_size
        ).inputs

        return self.predict_sample_with_fixed_blocking(
            sample,
            input_block_shape=input_block_shape,
            skip_preprocessing=skip_preprocessing,
            skip_postprocessing=skip_postprocessing,
        )

    # def predict(
    #     self,
    #     inputs: Predict_IO,
    #     skip_preprocessing: bool = False,
    #     skip_postprocessing: bool = False,
    # ) -> Predict_IO:
    #     """Run model prediction **including** pre/postprocessing."""

    #     if isinstance(inputs, Sample):
    #         return self.predict_sample_with_blocking(
    #             inputs,
    #             skip_preprocessing=skip_preprocessing,
    #             skip_postprocessing=skip_postprocessing,
    #         )
    #     elif isinstance(inputs, collections.abc.Iterable):
    #         return (
    #             self.predict(
    #                 ipt,
    #                 skip_preprocessing=skip_preprocessing,
    #                 skip_postprocessing=skip_postprocessing,
    #             )
    #             for ipt in inputs
    #         )
    #     else:
    #         assert_never(inputs)

    def apply_preprocessing(self, sample: Union[Sample, SampleBlockWithOrigin]) -> None:
        """apply preprocessing in-place, also updates sample stats"""
        for op in self._preprocessing:
            op(sample)

    def apply_postprocessing(
        self, sample: Union[Sample, SampleBlock, SampleBlockWithOrigin]
    ) -> None:
        """apply postprocessing in-place, also updates samples stats"""
        for op in self._postprocessing:
            if isinstance(sample, (Sample, SampleBlockWithOrigin)):
                op(sample)
            elif not isinstance(op, BlockedOperator):
                raise NotImplementedError(
                    "block wise update of output statistics not yet implemented"
                )
            else:
                op(sample)

    def load(self):
        """
        optional step: load model onto devices before calling forward if not using it as context manager
        """
        pass

    def unload(self):
        """
        free any device memory in use
        """
        self._adapter.unload()


def create_prediction_pipeline(
    bioimageio_model: AnyModelDescr,
    *,
    devices: Optional[Sequence[str]] = None,
    weight_format: Optional[WeightsFormat] = None,
    weights_format: Optional[WeightsFormat] = None,
    dataset_for_initial_statistics: Iterable[Union[Sample, Sequence[Tensor]]] = tuple(),
    keep_updating_initial_dataset_statistics: bool = False,
    fixed_dataset_statistics: Mapping[DatasetMeasure, MeasureValue] = MappingProxyType(
        {}
    ),
    model_adapter: Optional[ModelAdapter] = None,
    ns: Union[
        v0_5.ParameterizedSize_N,
        Mapping[Tuple[MemberId, AxisId], v0_5.ParameterizedSize_N],
    ] = 10,
    **deprecated_kwargs: Any,
) -> PredictionPipeline:
    """
    Creates prediction pipeline which includes:
    * computation of input statistics
    * preprocessing
    * model prediction
    * computation of output statistics
    * postprocessing
    """
    weights_format = weight_format or weights_format
    del weight_format
    if deprecated_kwargs:
        warnings.warn(
            f"deprecated create_prediction_pipeline kwargs: {set(deprecated_kwargs)}"
        )

    model_adapter = model_adapter or create_model_adapter(
        model_description=bioimageio_model,
        devices=devices,
        weight_format_priority_order=weights_format and (weights_format,),
    )

    input_ids = get_member_ids(bioimageio_model.inputs)

    def dataset():
        common_stat: Stat = {}
        for i, x in enumerate(dataset_for_initial_statistics):
            if isinstance(x, Sample):
                yield x
            else:
                yield Sample(members=dict(zip(input_ids, x)), stat=common_stat, id=i)

    preprocessing, postprocessing = setup_pre_and_postprocessing(
        bioimageio_model,
        dataset(),
        keep_updating_initial_dataset_stats=keep_updating_initial_dataset_statistics,
        fixed_dataset_stats=fixed_dataset_statistics,
    )

    return PredictionPipeline(
        name=bioimageio_model.name,
        model_description=bioimageio_model,
        model_adapter=model_adapter,
        preprocessing=preprocessing,
        postprocessing=postprocessing,
        default_ns=ns,
    )



import warnings
from typing import Any, List, Optional, Sequence, Union

from numpy.typing import NDArray

from bioimageio.spec.model import v0_4, v0_5
from bioimageio.spec.utils import download

from ..digest_spec import get_axes_infos
from ..tensor import Tensor
from ._model_adapter import ModelAdapter

try:
    import onnxruntime as rt
except Exception as e:
    rt = None
    rt_error = str(e)
else:
    rt_error = None


class ONNXModelAdapter(ModelAdapter):
    def __init__(
        self,
        *,
        model_description: Union[v0_4.ModelDescr, v0_5.ModelDescr],
        devices: Optional[Sequence[str]] = None,
    ):
        if rt is None:
            raise ImportError(f"failed to import onnxruntime: {rt_error}")

        super().__init__()
        self._internal_output_axes = [
            tuple(a.id for a in get_axes_infos(out))
            for out in model_description.outputs
        ]
        if model_description.weights.onnx is None:
            raise ValueError("No ONNX weights specified for {model_description.name}")

        self._session = rt.InferenceSession(
            str(download(model_description.weights.onnx.source).path)
        )
        onnx_inputs = self._session.get_inputs()  # type: ignore
        self._input_names: List[str] = [ipt.name for ipt in onnx_inputs]  # type: ignore

        if devices is not None:
            warnings.warn(
                f"Device management is not implemented for onnx yet, ignoring the devices {devices}"
            )

    def forward(self, *input_tensors: Optional[Tensor]) -> List[Optional[Tensor]]:
        assert len(input_tensors) == len(self._input_names)
        input_arrays = [None if ipt is None else ipt.data.data for ipt in input_tensors]
        result: Union[Sequence[Optional[NDArray[Any]]], Optional[NDArray[Any]]]
        result = self._session.run(  # pyright: ignore[reportUnknownVariableType]
            None, dict(zip(self._input_names, input_arrays))
        )
        if isinstance(result, (list, tuple)):
            result_seq: Sequence[Optional[NDArray[Any]]] = result
        else:
            result_seq = [result]  # type: ignore

        return [
            None if r is None else Tensor(r, dims=axes)
            for r, axes in zip(result_seq, self._internal_output_axes)
        ]

    def unload(self) -> None:
        warnings.warn(
            "Device management is not implemented for onnx yet, cannot unload model"
        )



"""bioimageio CLI

Note: Some docstrings use a hair space ' '
      to place the added '(default: ...)' on a new line.
"""

import json
import shutil
import subprocess
import sys
from argparse import RawTextHelpFormatter
from difflib import SequenceMatcher
from functools import cached_property
from pathlib import Path
from pprint import pformat, pprint
from typing import (
    Any,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    Union,
)

from loguru import logger
from pydantic import BaseModel, model_validator
from pydantic_settings import (
    BaseSettings,
    CliPositionalArg,
    CliSettingsSource,
    CliSubCommand,
    JsonConfigSettingsSource,
    PydanticBaseSettingsSource,
    SettingsConfigDict,
    YamlConfigSettingsSource,
)
from ruyaml import YAML
from tqdm import tqdm

from bioimageio.core import (
    MemberId,
    Sample,
    __version__,
    create_prediction_pipeline,
)
from bioimageio.core.commands import (
    WeightFormatArgAll,
    WeightFormatArgAny,
    package,
    test,
    validate_format,
)
from bioimageio.core.common import SampleId
from bioimageio.core.digest_spec import get_member_ids, load_sample_for_model
from bioimageio.core.io import load_dataset_stat, save_dataset_stat, save_sample
from bioimageio.core.proc_setup import (
    DatasetMeasure,
    Measure,
    MeasureValue,
    StatsCalculator,
    get_required_dataset_measures,
)
from bioimageio.core.stat_measures import Stat
from bioimageio.spec import (
    AnyModelDescr,
    InvalidDescr,
    load_description,
)
from bioimageio.spec._internal.types import NotEmpty
from bioimageio.spec.dataset import DatasetDescr
from bioimageio.spec.model import ModelDescr, v0_4, v0_5
from bioimageio.spec.notebook import NotebookDescr
from bioimageio.spec.utils import download, ensure_description_is_model

yaml = YAML(typ="safe")


class CmdBase(BaseModel, use_attribute_docstrings=True):
    pass


class ArgMixin(BaseModel, use_attribute_docstrings=True):
    pass


class WithSource(ArgMixin):
    source: CliPositionalArg[str]
    """Url/path to a bioimageio.yaml/rdf.yaml file
    or a bioimage.io resource identifier, e.g. 'affable-shark'"""

    @cached_property
    def descr(self):
        return load_description(self.source)

    @property
    def descr_id(self) -> str:
        """a more user-friendly description id
        (replacing legacy ids with their nicknames)
        """
        if isinstance(self.descr, InvalidDescr):
            return str(getattr(self.descr, "id", getattr(self.descr, "name")))
        else:
            return str(
                (
                    (bio_config := self.descr.config.get("bioimageio", {}))
                    and isinstance(bio_config, dict)
                    and bio_config.get("nickname")
                )
                or self.descr.id
                or self.descr.name
            )


class ValidateFormatCmd(CmdBase, WithSource):
    """bioimageio-validate-format - validate the meta data format of a bioimageio resource."""

    def run(self):
        validate_format(self.descr)


class TestCmd(CmdBase, WithSource):
    """bioimageio-test - Test a bioimageio resource (beyond meta data formatting)"""

    weight_format: WeightFormatArgAll = "all"
    """The weight format to limit testing to.

    (only relevant for model resources)"""

    devices: Optional[Union[str, Sequence[str]]] = None
    """Device(s) to use for testing"""

    decimal: int = 4
    """Precision for numerical comparisons"""

    def run(self):
        test(
            self.descr,
            weight_format=self.weight_format,
            devices=self.devices,
            decimal=self.decimal,
        )


class PackageCmd(CmdBase, WithSource):
    """bioimageio-package - save a resource's metadata with its associated files."""

    path: CliPositionalArg[Path]
    """The path to write the (zipped) package to.
    If it does not have a `.zip` suffix
    this command will save the package as an unzipped folder instead."""

    weight_format: WeightFormatArgAll = "all"
    """The weight format to include in the package (for model descriptions only)."""

    def run(self):
        if isinstance(self.descr, InvalidDescr):
            self.descr.validation_summary.display()
            raise ValueError("resource description is invalid")

        package(
            self.descr,
            self.path,
            weight_format=self.weight_format,
        )


def _get_stat(
    model_descr: AnyModelDescr,
    dataset: Iterable[Sample],
    dataset_length: int,
    stats_path: Path,
) -> Mapping[DatasetMeasure, MeasureValue]:
    req_dataset_meas, _ = get_required_dataset_measures(model_descr)
    if not req_dataset_meas:
        return {}

    req_dataset_meas, _ = get_required_dataset_measures(model_descr)

    if stats_path.exists():
        logger.info(f"loading precomputed dataset measures from {stats_path}")
        stat = load_dataset_stat(stats_path)
        for m in req_dataset_meas:
            if m not in stat:
                raise ValueError(f"Missing {m} in {stats_path}")

        return stat

    stats_calc = StatsCalculator(req_dataset_meas)

    for sample in tqdm(
        dataset, total=dataset_length, desc="precomputing dataset stats", unit="sample"
    ):
        stats_calc.update(sample)

    stat = stats_calc.finalize()
    save_dataset_stat(stat, stats_path)

    return stat


class PredictCmd(CmdBase, WithSource):
    """bioimageio-predict - Run inference on your data with a bioimage.io model."""

    inputs: NotEmpty[Sequence[Union[str, NotEmpty[Tuple[str, ...]]]]] = (
        "{input_id}/001.tif",
    )
    """Model input sample paths (for each input tensor)

    The input paths are expected to have shape...
     - (n_samples,) or (n_samples,1) for models expecting a single input tensor
     - (n_samples,) containing the substring '{input_id}', or
     - (n_samples, n_model_inputs) to provide each input tensor path explicitly.

    All substrings that are replaced by metadata from the model description:
    - '{model_id}'
    - '{input_id}'

    Example inputs to process sample 'a' and 'b'
    for a model expecting a 'raw' and a 'mask' input tensor:
    --inputs="[[\"a_raw.tif\",\"a_mask.tif\"],[\"b_raw.tif\",\"b_mask.tif\"]]"
    (Note that JSON double quotes need to be escaped.)

    Alternatively a `bioimageio-cli.yaml` (or `bioimageio-cli.json`) file
    may provide the arguments, e.g.:
    ```yaml
    inputs:
    - [a_raw.tif, a_mask.tif]
    - [b_raw.tif, b_mask.tif]
    ```

    `.npy` and any file extension supported by imageio are supported.
     Aavailable formats are listed at
    https://imageio.readthedocs.io/en/stable/formats/index.html#all-formats.
    Some formats have additional dependencies.

     
    """

    outputs: Union[str, NotEmpty[Tuple[str, ...]]] = (
        "outputs_{model_id}/{output_id}/{sample_id}.tif"
    )
    """Model output path pattern (per output tensor)

    All substrings that are replaced:
    - '{model_id}' (from model description)
    - '{output_id}' (from model description)
    - '{sample_id}' (extracted from input paths)

     
    """

    overwrite: bool = False
    """allow overwriting existing output files"""

    blockwise: bool = False
    """process inputs blockwise"""

    stats: Path = Path("dataset_statistics.json")
    """path to dataset statistics
    (will be written if it does not exist,
    but the model requires statistical dataset measures)
     """

    preview: bool = False
    """preview which files would be processed
    and what outputs would be generated."""

    weight_format: WeightFormatArgAny = "any"
    """The weight format to use."""

    example: bool = False
    """generate and run an example

    1. downloads example model inputs
    2. creates a `{model_id}_example` folder
    3. writes input arguments to `{model_id}_example/bioimageio-cli.yaml`
    4. executes a preview dry-run
    5. executes prediction with example input

     
    """

    def _example(self):
        model_descr = ensure_description_is_model(self.descr)
        input_ids = get_member_ids(model_descr.inputs)
        example_inputs = (
            model_descr.sample_inputs
            if isinstance(model_descr, v0_4.ModelDescr)
            else [ipt.sample_tensor or ipt.test_tensor for ipt in model_descr.inputs]
        )
        if not example_inputs:
            raise ValueError(f"{self.descr_id} does not specify any example inputs.")

        inputs001: List[str] = []
        example_path = Path(f"{self.descr_id}_example")
        example_path.mkdir(exist_ok=True)

        for t, src in zip(input_ids, example_inputs):
            local = download(src).path
            dst = Path(f"{example_path}/{t}/001{''.join(local.suffixes)}")
            dst.parent.mkdir(parents=True, exist_ok=True)
            inputs001.append(dst.as_posix())
            shutil.copy(local, dst)

        inputs = [tuple(inputs001)]
        output_pattern = f"{example_path}/outputs/{{output_id}}/{{sample_id}}.tif"

        bioimageio_cli_path = example_path / YAML_FILE
        stats_file = "dataset_statistics.json"
        stats = (example_path / stats_file).as_posix()
        yaml.dump(
            dict(
                inputs=inputs,
                outputs=output_pattern,
                stats=stats_file,
                blockwise=self.blockwise,
            ),
            bioimageio_cli_path,
        )

        yaml_file_content = None

        # escaped double quotes
        inputs_json = json.dumps(inputs)
        inputs_escaped = inputs_json.replace('"', r"\"")
        source_escaped = self.source.replace('"', r"\"")

        def get_example_command(preview: bool, escape: bool = False):
            q: str = '"' if escape else ""

            return [
                "bioimageio",
                "predict",
                f"--preview={preview}",  # update once we use implicit flags, see `class Bioimageio` below
                "--overwrite=True",
                f"--blockwise={self.blockwise}",
                f"--stats={q}{stats}{q}",
                f"--inputs={q}{inputs_escaped if escape else inputs_json}{q}",
                f"--outputs={q}{output_pattern}{q}",
                f"{q}{source_escaped if escape else self.source}{q}",
            ]

        if Path(YAML_FILE).exists():
            logger.info(
                "temporarily removing '{}' to execute example prediction", YAML_FILE
            )
            yaml_file_content = Path(YAML_FILE).read_bytes()
            Path(YAML_FILE).unlink()

        try:
            _ = subprocess.run(get_example_command(True), check=True)
            _ = subprocess.run(get_example_command(False), check=True)
        finally:
            if yaml_file_content is not None:
                _ = Path(YAML_FILE).write_bytes(yaml_file_content)
                logger.debug("restored '{}'", YAML_FILE)

        print(
            "🎉 Sucessfully ran example prediction!\n"
            + "To predict the example input using the CLI example config file"
            + f" {example_path/YAML_FILE}, execute `bioimageio predict` from {example_path}:\n"
            + f"$ cd {str(example_path)}\n"
            + f'$ bioimageio predict "{source_escaped}"\n\n'
            + "Alternatively run the following command"
            + " in the current workind directory, not the example folder:\n$ "
            + " ".join(get_example_command(False, escape=True))
            + f"\n(note that a local '{JSON_FILE}' or '{YAML_FILE}' may interfere with this)"
        )

    def run(self):
        if self.example:
            return self._example()

        model_descr = ensure_description_is_model(self.descr)

        input_ids = get_member_ids(model_descr.inputs)
        output_ids = get_member_ids(model_descr.outputs)

        minimum_input_ids = tuple(
            str(ipt.id) if isinstance(ipt, v0_5.InputTensorDescr) else str(ipt.name)
            for ipt in model_descr.inputs
            if not isinstance(ipt, v0_5.InputTensorDescr) or not ipt.optional
        )
        maximum_input_ids = tuple(
            str(ipt.id) if isinstance(ipt, v0_5.InputTensorDescr) else str(ipt.name)
            for ipt in model_descr.inputs
        )

        def expand_inputs(i: int, ipt: Union[str, Tuple[str, ...]]) -> Tuple[str, ...]:
            if isinstance(ipt, str):
                ipts = tuple(
                    ipt.format(model_id=self.descr_id, input_id=t) for t in input_ids
                )
            else:
                ipts = tuple(
                    p.format(model_id=self.descr_id, input_id=t)
                    for t, p in zip(input_ids, ipt)
                )

            if len(set(ipts)) < len(ipts):
                if len(minimum_input_ids) == len(maximum_input_ids):
                    n = len(minimum_input_ids)
                else:
                    n = f"{len(minimum_input_ids)}-{len(maximum_input_ids)}"

                raise ValueError(
                    f"[input sample #{i}] Include '{{input_id}}' in path pattern or explicitly specify {n} distinct input paths (got {ipt})"
                )

            if len(ipts) < len(minimum_input_ids):
                raise ValueError(
                    f"[input sample #{i}] Expected at least {len(minimum_input_ids)} inputs {minimum_input_ids}, got {ipts}"
                )

            if len(ipts) > len(maximum_input_ids):
                raise ValueError(
                    f"Expected at most {len(maximum_input_ids)} inputs {maximum_input_ids}, got {ipts}"
                )

            return ipts

        inputs = [expand_inputs(i, ipt) for i, ipt in enumerate(self.inputs, start=1)]

        sample_paths_in = [
            {t: Path(p) for t, p in zip(input_ids, ipts)} for ipts in inputs
        ]

        sample_ids = _get_sample_ids(sample_paths_in)

        def expand_outputs():
            if isinstance(self.outputs, str):
                outputs = [
                    tuple(
                        Path(
                            self.outputs.format(
                                model_id=self.descr_id, output_id=t, sample_id=s
                            )
                        )
                        for t in output_ids
                    )
                    for s in sample_ids
                ]
            else:
                outputs = [
                    tuple(
                        Path(p.format(model_id=self.descr_id, output_id=t, sample_id=s))
                        for t, p in zip(output_ids, self.outputs)
                    )
                    for s in sample_ids
                ]

            for i, out in enumerate(outputs, start=1):
                if len(set(out)) < len(out):
                    raise ValueError(
                        f"[output sample #{i}] Include '{{output_id}}' in path pattern or explicitly specify {len(output_ids)} distinct output paths (got {out})"
                    )

                if len(out) != len(output_ids):
                    raise ValueError(
                        f"[output sample #{i}] Expected {len(output_ids)} outputs {output_ids}, got {out}"
                    )

            return outputs

        outputs = expand_outputs()

        sample_paths_out = [
            {MemberId(t): Path(p) for t, p in zip(output_ids, out)} for out in outputs
        ]

        if not self.overwrite:
            for sample_paths in sample_paths_out:
                for p in sample_paths.values():
                    if p.exists():
                        raise FileExistsError(
                            f"{p} already exists. use --overwrite to (re-)write outputs anyway."
                        )
        if self.preview:
            print("🛈 bioimageio prediction preview structure:")
            pprint(
                {
                    "{sample_id}": dict(
                        inputs={"{input_id}": "<input path>"},
                        outputs={"{output_id}": "<output path>"},
                    )
                }
            )
            print("🔎 bioimageio prediction preview output:")
            pprint(
                {
                    s: dict(
                        inputs={t: p.as_posix() for t, p in sp_in.items()},
                        outputs={t: p.as_posix() for t, p in sp_out.items()},
                    )
                    for s, sp_in, sp_out in zip(
                        sample_ids, sample_paths_in, sample_paths_out
                    )
                }
            )
            return

        def input_dataset(stat: Stat):
            for s, sp_in in zip(sample_ids, sample_paths_in):
                yield load_sample_for_model(
                    model=model_descr,
                    paths=sp_in,
                    stat=stat,
                    sample_id=s,
                )

        stat: Dict[Measure, MeasureValue] = dict(
            _get_stat(
                model_descr, input_dataset({}), len(sample_ids), self.stats
            ).items()
        )

        pp = create_prediction_pipeline(
            model_descr,
            weight_format=None if self.weight_format == "any" else self.weight_format,
        )
        predict_method = (
            pp.predict_sample_with_blocking
            if self.blockwise
            else pp.predict_sample_without_blocking
        )

        for sample_in, sp_out in tqdm(
            zip(input_dataset(dict(stat)), sample_paths_out),
            total=len(inputs),
            desc=f"predict with {self.descr_id}",
            unit="sample",
        ):
            sample_out = predict_method(sample_in)
            save_sample(sp_out, sample_out)


JSON_FILE = "bioimageio-cli.json"
YAML_FILE = "bioimageio-cli.yaml"


class Bioimageio(
    BaseSettings,
    # alias_generator=AliasGenerator(
    #     validation_alias=lambda s: AliasChoices(s, to_snake(s).replace("_", "-"))
    # ),
    # TODO: investigate how to allow a validation alias for subcommands
    #       ('validate-format' vs 'validate_format')
    cli_parse_args=True,
    cli_prog_name="bioimageio",
    cli_use_class_docs_for_groups=True,
    # cli_implicit_flags=True, # TODO: make flags implicit, see https://github.com/pydantic/pydantic-settings/issues/361
    use_attribute_docstrings=True,
):
    """bioimageio - CLI for bioimage.io resources 🦒"""

    model_config = SettingsConfigDict(json_file=JSON_FILE, yaml_file=YAML_FILE)

    validate_format: CliSubCommand[ValidateFormatCmd]
    "Check a resource's metadata format"

    test: CliSubCommand[TestCmd]
    "Test a bioimageio resource (beyond meta data formatting)"

    package: CliSubCommand[PackageCmd]
    "Package a resource"

    predict: CliSubCommand[PredictCmd]
    "Predict with a model resource"

    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: Type[BaseSettings],
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,
        dotenv_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
    ) -> Tuple[PydanticBaseSettingsSource, ...]:
        cli: CliSettingsSource[BaseSettings] = CliSettingsSource(
            settings_cls,
            cli_parse_args=True,
            formatter_class=RawTextHelpFormatter,
        )
        sys_args = pformat(sys.argv)
        logger.info("starting CLI with arguments:\n{}", sys_args)
        return (
            cli,
            init_settings,
            YamlConfigSettingsSource(settings_cls),
            JsonConfigSettingsSource(settings_cls),
        )

    @model_validator(mode="before")
    @classmethod
    def _log(cls, data: Any):
        logger.info(
            "loaded CLI input:\n{}",
            pformat({k: v for k, v in data.items() if v is not None}),
        )
        return data

    def run(self):
        logger.info(
            "executing CLI command:\n{}",
            pformat({k: v for k, v in self.model_dump().items() if v is not None}),
        )
        cmd = self.validate_format or self.test or self.package or self.predict
        assert cmd is not None
        cmd.run()


assert isinstance(Bioimageio.__doc__, str)
Bioimageio.__doc__ += f"""

library versions:
  bioimageio.core {__version__}
  bioimageio.spec {__version__}

spec format versions:
        model RDF {ModelDescr.implemented_format_version}
      dataset RDF {DatasetDescr.implemented_format_version}
     notebook RDF {NotebookDescr.implemented_format_version}

"""


def _get_sample_ids(
    input_paths: Sequence[Mapping[MemberId, Path]]
) -> Sequence[SampleId]:
    """Get sample ids for given input paths, based on the common path per sample.

    Falls back to sample01, samle02, etc..."""

    matcher = SequenceMatcher()

    def get_common_seq(seqs: Sequence[Sequence[str]]) -> Sequence[str]:
        """extract a common sequence from multiple sequences
        (order sensitive; strips whitespace and slashes)
        """
        common = seqs[0]

        for seq in seqs[1:]:
            if not seq:
                continue
            matcher.set_seqs(common, seq)
            i, _, size = matcher.find_longest_match()
            common = common[i : i + size]

        if isinstance(common, str):
            common = common.strip().strip("/")
        else:
            common = [cs for c in common if (cs := c.strip().strip("/"))]

        if not common:
            raise ValueError(f"failed to find common sequence for {seqs}")

        return common

    def get_shorter_diff(seqs: Sequence[Sequence[str]]) -> List[Sequence[str]]:
        """get a shorter sequence whose entries are still unique
        (order sensitive, not minimal sequence)
        """
        min_seq_len = min(len(s) for s in seqs)
        # cut from the start
        for start in range(min_seq_len - 1, -1, -1):
            shortened = [s[start:] for s in seqs]
            if len(set(shortened)) == len(seqs):
                min_seq_len -= start
                break
        else:
            seen: Set[Sequence[str]] = set()
            dupes = [s for s in seqs if s in seen or seen.add(s)]
            raise ValueError(f"Found duplicate entries {dupes}")

        # cut from the end
        for end in range(min_seq_len - 1, 1, -1):
            shortened = [s[:end] for s in shortened]
            if len(set(shortened)) == len(seqs):
                break

        return shortened

    full_tensor_ids = [
        sorted(
            p.resolve().with_suffix("").as_posix() for p in input_sample_paths.values()
        )
        for input_sample_paths in input_paths
    ]
    try:
        long_sample_ids = [get_common_seq(t) for t in full_tensor_ids]
        sample_ids = get_shorter_diff(long_sample_ids)
    except ValueError as e:
        raise ValueError(f"failed to extract sample ids: {e}")

    return sample_ids



import collections.abc
from os import PathLike
from pathlib import Path
from typing import Any, Mapping, Optional, Sequence, Union

import imageio
from imageio.v3 import imread, imwrite
from loguru import logger
from numpy.typing import NDArray
from pydantic import BaseModel, ConfigDict, TypeAdapter

from bioimageio.core.common import PerMember
from bioimageio.core.stat_measures import DatasetMeasure, MeasureValue
from bioimageio.spec.utils import load_array, save_array

from .axis import Axis, AxisLike
from .sample import Sample
from .tensor import Tensor


def load_image(path: Path, is_volume: Optional[bool] = None) -> NDArray[Any]:
    """load a single image as numpy array

    Args:
        path: image path
        is_volume: deprecated
    """
    ext = path.suffix
    if ext == ".npy":
        return load_array(path)
    else:
        return imread(path)  # pyright: ignore[reportUnknownVariableType]


def load_tensor(path: Path, axes: Optional[Sequence[AxisLike]] = None) -> Tensor:
    # TODO: load axis meta data
    array = load_image(path)

    return Tensor.from_numpy(array, dims=axes)


def save_tensor(path: Path, tensor: Tensor) -> None:
    # TODO: save axis meta data

    data: NDArray[Any] = tensor.data.to_numpy()
    path = Path(path)
    path.parent.mkdir(exist_ok=True, parents=True)
    if path.suffix == ".npy":
        save_array(path, data)
    else:
        # if singleton_axes := [a for a, s in tensor.tagged_shape.items() if s == 1]:
        #     tensor = tensor[{a: 0 for a in singleton_axes}]
        #     singleton_axes_msg = f"(without singleton axes {singleton_axes}) "
        # else:
        singleton_axes_msg = ""

        logger.debug(
            "writing tensor {} {}to {}",
            dict(tensor.tagged_shape),
            singleton_axes_msg,
            path,
        )
        imwrite(path, data)


def save_sample(path: Union[Path, str, PerMember[Path]], sample: Sample) -> None:
    """save a sample to path

    If `path` is a pathlib.Path or a string it must contain `{member_id}` and may contain `{sample_id}`,
    which are resolved with the `sample` object.
    """

    if not isinstance(path, collections.abc.Mapping) and "{member_id}" not in str(path):
        raise ValueError(f"missing `{{member_id}}` in path {path}")

    for m, t in sample.members.items():
        if isinstance(path, collections.abc.Mapping):
            p = path[m]
        else:
            p = Path(str(path).format(sample_id=sample.id, member_id=m))

        save_tensor(p, t)


class _SerializedDatasetStatsEntry(
    BaseModel, frozen=True, arbitrary_types_allowed=True
):
    measure: DatasetMeasure
    value: MeasureValue


_stat_adapter = TypeAdapter(
    Sequence[_SerializedDatasetStatsEntry],
    config=ConfigDict(arbitrary_types_allowed=True),
)


def save_dataset_stat(stat: Mapping[DatasetMeasure, MeasureValue], path: Path):
    serializable = [
        _SerializedDatasetStatsEntry(measure=k, value=v) for k, v in stat.items()
    ]
    _ = path.write_bytes(_stat_adapter.dump_json(serializable))


def load_dataset_stat(path: Path):
    seq = _stat_adapter.validate_json(path.read_bytes())
    return {e.measure: e.value for e in seq}



from typing import (
    Iterable,
    List,
    Mapping,
    NamedTuple,
    Optional,
    Sequence,
    Set,
    Union,
)

from typing_extensions import assert_never

from bioimageio.core.common import MemberId
from bioimageio.core.digest_spec import get_member_ids
from bioimageio.spec.model import AnyModelDescr, v0_4, v0_5
from bioimageio.spec.model.v0_5 import TensorId

from .proc_ops import (
    AddKnownDatasetStats,
    EnsureDtype,
    Processing,
    UpdateStats,
    get_proc_class,
)
from .sample import Sample
from .stat_calculators import StatsCalculator
from .stat_measures import (
    DatasetMeasure,
    DatasetMeasureBase,
    Measure,
    MeasureValue,
    SampleMeasure,
    SampleMeasureBase,
)

TensorDescr = Union[
    v0_4.InputTensorDescr,
    v0_4.OutputTensorDescr,
    v0_5.InputTensorDescr,
    v0_5.OutputTensorDescr,
]


class PreAndPostprocessing(NamedTuple):
    pre: List[Processing]
    post: List[Processing]


class _SetupProcessing(NamedTuple):
    pre: List[Processing]
    post: List[Processing]
    pre_measures: Set[Measure]
    post_measures: Set[Measure]


def setup_pre_and_postprocessing(
    model: AnyModelDescr,
    dataset_for_initial_statistics: Iterable[Sample],
    keep_updating_initial_dataset_stats: bool = False,
    fixed_dataset_stats: Optional[Mapping[DatasetMeasure, MeasureValue]] = None,
) -> PreAndPostprocessing:
    """
    Get pre- and postprocessing operators for a `model` description.
    userd in `bioimageio.core.create_prediction_pipeline"""
    prep, post, prep_meas, post_meas = _prepare_setup_pre_and_postprocessing(model)

    missing_dataset_stats = {
        m
        for m in prep_meas | post_meas
        if fixed_dataset_stats is None or m not in fixed_dataset_stats
    }
    if missing_dataset_stats:
        initial_stats_calc = StatsCalculator(missing_dataset_stats)
        for sample in dataset_for_initial_statistics:
            initial_stats_calc.update(sample)

        initial_stats = initial_stats_calc.finalize()
    else:
        initial_stats = {}

    prep.insert(
        0,
        UpdateStats(
            StatsCalculator(prep_meas, initial_stats),
            keep_updating_initial_dataset_stats=keep_updating_initial_dataset_stats,
        ),
    )
    if post_meas:
        post.insert(
            0,
            UpdateStats(
                StatsCalculator(post_meas, initial_stats),
                keep_updating_initial_dataset_stats=keep_updating_initial_dataset_stats,
            ),
        )

    if fixed_dataset_stats:
        prep.insert(0, AddKnownDatasetStats(fixed_dataset_stats))
        post.insert(0, AddKnownDatasetStats(fixed_dataset_stats))

    return PreAndPostprocessing(prep, post)


class RequiredMeasures(NamedTuple):
    pre: Set[Measure]
    post: Set[Measure]


class RequiredDatasetMeasures(NamedTuple):
    pre: Set[DatasetMeasure]
    post: Set[DatasetMeasure]


class RequiredSampleMeasures(NamedTuple):
    pre: Set[SampleMeasure]
    post: Set[SampleMeasure]


def get_requried_measures(model: AnyModelDescr) -> RequiredMeasures:
    s = _prepare_setup_pre_and_postprocessing(model)
    return RequiredMeasures(s.pre_measures, s.post_measures)


def get_required_dataset_measures(model: AnyModelDescr) -> RequiredDatasetMeasures:
    s = _prepare_setup_pre_and_postprocessing(model)
    return RequiredDatasetMeasures(
        {m for m in s.pre_measures if isinstance(m, DatasetMeasureBase)},
        {m for m in s.post_measures if isinstance(m, DatasetMeasureBase)},
    )


def get_requried_sample_measures(model: AnyModelDescr) -> RequiredSampleMeasures:
    s = _prepare_setup_pre_and_postprocessing(model)
    return RequiredSampleMeasures(
        {m for m in s.pre_measures if isinstance(m, SampleMeasureBase)},
        {m for m in s.post_measures if isinstance(m, SampleMeasureBase)},
    )


def _prepare_setup_pre_and_postprocessing(model: AnyModelDescr) -> _SetupProcessing:
    pre_measures: Set[Measure] = set()
    post_measures: Set[Measure] = set()

    input_ids = set(get_member_ids(model.inputs))
    output_ids = set(get_member_ids(model.outputs))

    def prepare_procs(tensor_descrs: Sequence[TensorDescr]):
        procs: List[Processing] = []
        for t_descr in tensor_descrs:
            if isinstance(t_descr, (v0_4.InputTensorDescr, v0_5.InputTensorDescr)):
                proc_descrs: List[
                    Union[
                        v0_4.PreprocessingDescr,
                        v0_5.PreprocessingDescr,
                        v0_4.PostprocessingDescr,
                        v0_5.PostprocessingDescr,
                    ]
                ] = list(t_descr.preprocessing)
            elif isinstance(
                t_descr,
                (v0_4.OutputTensorDescr, v0_5.OutputTensorDescr),
            ):
                proc_descrs = list(t_descr.postprocessing)
            else:
                assert_never(t_descr)

            if isinstance(t_descr, (v0_4.InputTensorDescr, v0_4.OutputTensorDescr)):
                ensure_dtype = v0_5.EnsureDtypeDescr(
                    kwargs=v0_5.EnsureDtypeKwargs(dtype=t_descr.data_type)
                )
                if isinstance(t_descr, v0_4.InputTensorDescr) and proc_descrs:
                    proc_descrs.insert(0, ensure_dtype)

                proc_descrs.append(ensure_dtype)

            for proc_d in proc_descrs:
                proc_class = get_proc_class(proc_d)
                member_id = (
                    TensorId(str(t_descr.name))
                    if isinstance(t_descr, v0_4.TensorDescrBase)
                    else t_descr.id
                )
                req = proc_class.from_proc_descr(
                    proc_d, member_id  # pyright: ignore[reportArgumentType]
                )
                for m in req.required_measures:
                    if m.member_id in input_ids:
                        pre_measures.add(m)
                    elif m.member_id in output_ids:
                        post_measures.add(m)
                    else:
                        raise ValueError("When to raise ")
                procs.append(req)
        return procs

    return _SetupProcessing(
        pre=prepare_procs(model.inputs),
        post=prepare_procs(model.outputs),
        pre_measures=pre_measures,
        post_measures=post_measures,
    )



# untested draft!
import subprocess
from os import chdir
from pathlib import Path


def run(prompt: str):
    _ = subprocess.run(prompt, check=True, capture_output=True)


if __name__ == "__main__":
    repo_dir = Path(__file__).parent.parent.parent
    cur_dir = Path().resolve()
    chdir(str(repo_dir))
    try:
        run("mamba env create --file core-bioimage-io/dev/env.yaml")
        run(
            "pip install --no-deps --config-settings editable_mode=compat -e spec-bioimage-io"
        )
        run(
            "pip install --no-deps --config-settings editable_mode=compat -e core-bioimage-io"
        )
    except Exception:
        chdir(cur_dir)
        raise



from typing import Tuple, Union

import numpy as np
import pytest
from xarray.testing import assert_allclose  # pyright: ignore[reportUnknownVariableType]

from bioimageio.core.axis import AxisId
from bioimageio.core.common import MemberId
from bioimageio.core.sample import Sample
from bioimageio.core.stat_calculators import MeanVarStdCalculator
from bioimageio.core.stat_measures import (
    DatasetMean,
    DatasetStd,
    DatasetVar,
)
from bioimageio.core.tensor import Tensor


def create_random_dataset(tid: MemberId, axes: Tuple[AxisId, ...]):
    n = 3
    sizes = list(range(n, len(axes) + n))
    data = np.asarray(np.random.rand(*sizes))
    ds = [
        Sample(members={tid: Tensor(data[i : i + 1], dims=axes)}, stat={}, id=None)
        for i in range(n)
    ]
    return Tensor(data, dims=axes), ds


@pytest.mark.parametrize(
    "axes",
    [
        None,
        ("x", "y"),
        ("channel", "y"),
    ],
)
def test_mean_var_std_calculator(axes: Union[None, str, Tuple[str, ...]]):
    tid = MemberId("tensor")
    axes = tuple(map(AxisId, ("batch", "channel", "x", "y")))
    data, ds = create_random_dataset(tid, axes)
    expected_mean = data.mean(axes)
    expected_var = data.var(axes)
    expected_std = data.std(axes)

    calc = MeanVarStdCalculator(tid, axes=axes)
    for s in ds:
        calc.update(s)

    actual = calc.finalize()
    actual_mean = actual[DatasetMean(member_id=tid, axes=axes)]
    actual_var = actual[DatasetVar(member_id=tid, axes=axes)]
    actual_std = actual[DatasetStd(member_id=tid, axes=axes)]

    assert_allclose(
        actual_mean if isinstance(actual_mean, (int, float)) else actual_mean.data,
        expected_mean.data,
    )
    assert_allclose(
        actual_var if isinstance(actual_var, (int, float)) else actual_var.data,
        expected_var.data,
    )
    assert_allclose(
        actual_std if isinstance(actual_std, (int, float)) else actual_std.data,
        expected_std.data,
    )



import subprocess
from pathlib import Path
from tempfile import TemporaryDirectory

import pooch

from bioimageio.core import load_description, save_bioimageio_yaml_only

if __name__ == "__main__":
    rdf_source = "https://raw.githubusercontent.com/bioimage-io/spec-bioimage-io/main/example_descriptions/models/unet2d_nuclei_broad/v0_4_9.bioimageio.yaml"

    local_source = Path(pooch.retrieve(rdf_source, None))  # type: ignore
    model_as_is = load_description(rdf_source, format_version="discover")
    model_latest = load_description(rdf_source, format_version="latest")
    print(model_latest.validation_summary)

    with TemporaryDirectory() as tmp:
        as_is = Path(tmp) / "as_is.bioimageio.yaml"

        save_bioimageio_yaml_only(
            model_as_is, file=as_is
        )  # write out as is to avoid sorting diff
        latest = Path(tmp) / "latest.bioimageio.yaml"
        save_bioimageio_yaml_only(model_latest, file=latest)

        _ = subprocess.run(f"git diff --no-index --ignore-all-space {as_is} {latest}")



import json
import sys
from pathlib import Path

if sys.version_info < (3, 9):

    def files(package_name: str):
        assert package_name == "bioimageio.core"
        return Path(__file__).parent.parent

else:
    from importlib.resources import files as files


with files("bioimageio.core").joinpath("VERSION").open("r", encoding="utf-8") as f:
    VERSION = json.load(f)["version"]
    assert isinstance(VERSION, str)



from __future__ import annotations

from dataclasses import dataclass
from typing import Literal, Mapping, Optional, TypeVar, Union

from typing_extensions import assert_never

from bioimageio.spec.model import v0_5


def _get_axis_type(a: Literal["b", "t", "i", "c", "x", "y", "z"]):
    if a == "b":
        return "batch"
    elif a == "t":
        return "time"
    elif a == "i":
        return "index"
    elif a == "c":
        return "channel"
    elif a in ("x", "y", "z"):
        return "space"
    else:
        return "index"  # return most unspecific axis


S = TypeVar("S", bound=str)


AxisId = v0_5.AxisId

T = TypeVar("T")
PerAxis = Mapping[AxisId, T]

BatchSize = int

AxisLetter = Literal["b", "i", "t", "c", "z", "y", "x"]
AxisLike = Union[AxisId, AxisLetter, v0_5.AnyAxis, "Axis"]


@dataclass
class Axis:
    id: AxisId
    type: Literal["batch", "channel", "index", "space", "time"]

    @classmethod
    def create(cls, axis: AxisLike) -> Axis:
        if isinstance(axis, cls):
            return axis
        elif isinstance(axis, Axis):
            return Axis(id=axis.id, type=axis.type)
        elif isinstance(axis, str):
            return Axis(id=AxisId(axis), type=_get_axis_type(axis))
        elif isinstance(axis, v0_5.AxisBase):
            return Axis(id=AxisId(axis.id), type=axis.type)
        else:
            assert_never(axis)


@dataclass
class AxisInfo(Axis):
    maybe_singleton: bool  # TODO: replace 'maybe_singleton' with size min/max for better axis guessing

    @classmethod
    def create(cls, axis: AxisLike, maybe_singleton: Optional[bool] = None) -> AxisInfo:
        if isinstance(axis, AxisInfo):
            return axis

        axis_base = super().create(axis)
        if maybe_singleton is None:
            if isinstance(axis, (Axis, str)):
                maybe_singleton = True
            else:
                if axis.size is None:
                    maybe_singleton = True
                elif isinstance(axis.size, int):
                    maybe_singleton = axis.size == 1
                elif isinstance(axis.size, v0_5.SizeReference):
                    maybe_singleton = (
                        True  # TODO: check if singleton is ok for a `SizeReference`
                    )
                elif isinstance(
                    axis.size, (v0_5.ParameterizedSize, v0_5.DataDependentSize)
                ):
                    try:
                        maybe_size_one = axis.size.validate_size(
                            1
                        )  # TODO: refactor validate_size() to have boolean func here
                    except ValueError:
                        maybe_singleton = False
                    else:
                        maybe_singleton = maybe_size_one == 1
                else:
                    assert_never(axis.size)

        return AxisInfo(
            id=axis_base.id, type=axis_base.type, maybe_singleton=maybe_singleton
        )



import warnings
from abc import ABC, abstractmethod
from typing import List, Optional, Sequence, Tuple, Union, final

from bioimageio.spec.model import v0_4, v0_5

from ..tensor import Tensor

WeightsFormat = Union[v0_4.WeightsFormat, v0_5.WeightsFormat]

# Known weight formats in order of priority
# First match wins
DEFAULT_WEIGHT_FORMAT_PRIORITY_ORDER: Tuple[WeightsFormat, ...] = (
    "pytorch_state_dict",
    "tensorflow_saved_model_bundle",
    "torchscript",
    "onnx",
    "keras_hdf5",
)


class ModelAdapter(ABC):
    """
    Represents model *without* any preprocessing or postprocessing.

    ```
    from bioimageio.core import load_description

    model = load_description(...)

    # option 1:
    adapter = ModelAdapter.create(model)
    adapter.forward(...)
    adapter.unload()

    # option 2:
    with ModelAdapter.create(model) as adapter:
        adapter.forward(...)
    ```
    """

    @final
    @classmethod
    def create(
        cls,
        model_description: Union[v0_4.ModelDescr, v0_5.ModelDescr],
        *,
        devices: Optional[Sequence[str]] = None,
        weight_format_priority_order: Optional[Sequence[WeightsFormat]] = None,
    ):
        """
        Creates model adapter based on the passed spec
        Note: All specific adapters should happen inside this function to prevent different framework
        initializations interfering with each other
        """
        if not isinstance(model_description, (v0_4.ModelDescr, v0_5.ModelDescr)):
            raise TypeError(
                f"expected v0_4.ModelDescr or v0_5.ModelDescr, but got {type(model_description)}"
            )

        weights = model_description.weights
        errors: List[Tuple[WeightsFormat, Exception]] = []
        weight_format_priority_order = (
            DEFAULT_WEIGHT_FORMAT_PRIORITY_ORDER
            if weight_format_priority_order is None
            else weight_format_priority_order
        )
        # limit weight formats to the ones present
        weight_format_priority_order = [
            w for w in weight_format_priority_order if getattr(weights, w) is not None
        ]

        for wf in weight_format_priority_order:
            if wf == "pytorch_state_dict" and weights.pytorch_state_dict is not None:
                try:
                    from ._pytorch_model_adapter import PytorchModelAdapter

                    return PytorchModelAdapter(
                        outputs=model_description.outputs,
                        weights=weights.pytorch_state_dict,
                        devices=devices,
                    )
                except Exception as e:
                    errors.append((wf, e))
            elif (
                wf == "tensorflow_saved_model_bundle"
                and weights.tensorflow_saved_model_bundle is not None
            ):
                try:
                    from ._tensorflow_model_adapter import TensorflowModelAdapter

                    return TensorflowModelAdapter(
                        model_description=model_description, devices=devices
                    )
                except Exception as e:
                    errors.append((wf, e))
            elif wf == "onnx" and weights.onnx is not None:
                try:
                    from ._onnx_model_adapter import ONNXModelAdapter

                    return ONNXModelAdapter(
                        model_description=model_description, devices=devices
                    )
                except Exception as e:
                    errors.append((wf, e))
            elif wf == "torchscript" and weights.torchscript is not None:
                try:
                    from ._torchscript_model_adapter import TorchscriptModelAdapter

                    return TorchscriptModelAdapter(
                        model_description=model_description, devices=devices
                    )
                except Exception as e:
                    errors.append((wf, e))
            elif wf == "keras_hdf5" and weights.keras_hdf5 is not None:
                # keras can either be installed as a separate package or used as part of tensorflow
                # we try to first import the keras model adapter using the separate package and,
                # if it is not available, try to load the one using tf
                try:
                    from ._keras_model_adapter import (
                        KerasModelAdapter,
                        keras,  # type: ignore
                    )

                    if keras is None:
                        from ._tensorflow_model_adapter import KerasModelAdapter

                    return KerasModelAdapter(
                        model_description=model_description, devices=devices
                    )
                except Exception as e:
                    errors.append((wf, e))

        assert errors
        if len(weight_format_priority_order) == 1:
            assert len(errors) == 1
            raise ValueError(
                f"The '{weight_format_priority_order[0]}' model adapter could not be created"
                + f" in this environment:\n{errors[0][1].__class__.__name__}({errors[0][1]}).\n\n"
            )

        else:
            error_list = "\n - ".join(
                f"{wf}: {e.__class__.__name__}({e})" for wf, e in errors
            )
            raise ValueError(
                "None of the weight format specific model adapters could be created"
                + f" in this environment. Errors are:\n\n{error_list}.\n\n"
            )

    @final
    def load(self, *, devices: Optional[Sequence[str]] = None) -> None:
        warnings.warn("Deprecated. ModelAdapter is loaded on initialization")

    @abstractmethod
    def forward(self, *input_tensors: Optional[Tensor]) -> List[Optional[Tensor]]:
        """
        Run forward pass of model to get model predictions
        """
        # TODO: handle tensor.transpose in here and make _forward_impl the abstract impl

    @abstractmethod
    def unload(self):
        """
        Unload model from any devices, freeing their memory.
        The moder adapter should be considered unusable afterwards.
        """


def get_weight_formats() -> List[str]:
    """
    Return list of supported weight types
    """
    return list(DEFAULT_WEIGHT_FORMAT_PRIORITY_ORDER)


create_model_adapter = ModelAdapter.create



# type: ignore  # TODO enable type checking
import zipfile
from pathlib import Path

import pytest

from bioimageio.spec import load_description
from bioimageio.spec.model.v0_5 import ModelDescr


@pytest.mark.skip(
    "tensorflow converter not updated yet"
)  # TODO: test tensorflow converter
def test_tensorflow_converter(any_keras_model: Path, tmp_path: Path):
    from bioimageio.core.weight_converter.keras import (
        convert_weights_to_tensorflow_saved_model_bundle,
    )

    out_path = tmp_path / "weights"
    model = load_description(any_keras_model)
    assert isinstance(model, ModelDescr), model.validation_summary.format()
    ret_val = convert_weights_to_tensorflow_saved_model_bundle(model, out_path)
    assert out_path.exists()
    assert (out_path / "variables").exists()
    assert (out_path / "saved_model.pb").exists()
    assert (
        ret_val == 0
    )  # check for correctness is done in converter and returns 0 if it passes


@pytest.mark.skip(
    "tensorflow converter not updated yet"
)  # TODO: test tensorflow converter
def test_tensorflow_converter_zipped(any_keras_model: Path, tmp_path: Path):
    from bioimageio.core.weight_converter.keras import (
        convert_weights_to_tensorflow_saved_model_bundle,
    )

    out_path = tmp_path / "weights.zip"
    model = load_description(any_keras_model)
    assert isinstance(model, ModelDescr), model.validation_summary.format()
    ret_val = convert_weights_to_tensorflow_saved_model_bundle(model, out_path)
    assert out_path.exists()
    assert (
        ret_val == 0
    )  # check for correctness is done in converter and returns 0 if it passes

    # make sure that the zip package was created correctly
    expected_names = {"saved_model.pb", "variables/variables.index"}
    with zipfile.ZipFile(out_path, "r") as f:
        names = set([name for name in f.namelist()])
    assert len(expected_names - names) == 0



from bioimageio.spec import load_description_and_validate_format_only


def test_model(any_model: str):
    summary = load_description_and_validate_format_only(any_model)
    assert summary.status == "passed", summary.format()



from __future__ import annotations

from dataclasses import dataclass
from math import ceil, floor
from typing import (
    Callable,
    Dict,
    Generic,
    Iterable,
    Optional,
    Tuple,
    TypeVar,
    Union,
)

import numpy as np
from typing_extensions import Self

from bioimageio.core.block import Block

from .axis import AxisId, PerAxis
from .block_meta import (
    BlockMeta,
    LinearAxisTransform,
    split_multiple_shapes_into_blocks,
)
from .common import (
    BlockIndex,
    Halo,
    HaloLike,
    MemberId,
    PadMode,
    PerMember,
    SampleId,
    SliceInfo,
    TotalNumberOfBlocks,
)
from .stat_measures import Stat
from .tensor import Tensor

# TODO: allow for lazy samples to read/write to disk


@dataclass
class Sample:
    """A dataset sample"""

    members: Dict[MemberId, Tensor]
    """the sample's tensors"""

    stat: Stat
    """sample and dataset statistics"""

    id: SampleId
    """identifier within the sample's dataset"""

    @property
    def shape(self) -> PerMember[PerAxis[int]]:
        return {tid: t.sizes for tid, t in self.members.items()}

    def split_into_blocks(
        self,
        block_shapes: PerMember[PerAxis[int]],
        halo: PerMember[PerAxis[HaloLike]],
        pad_mode: PadMode,
        broadcast: bool = False,
    ) -> Tuple[TotalNumberOfBlocks, Iterable[SampleBlockWithOrigin]]:
        assert not (
            missing := [m for m in block_shapes if m not in self.members]
        ), f"`block_shapes` specified for unknown members: {missing}"
        assert not (
            missing := [m for m in halo if m not in block_shapes]
        ), f"`halo` specified for members without `block_shape`: {missing}"

        n_blocks, blocks = split_multiple_shapes_into_blocks(
            shapes=self.shape,
            block_shapes=block_shapes,
            halo=halo,
            broadcast=broadcast,
        )
        return n_blocks, sample_block_generator(blocks, origin=self, pad_mode=pad_mode)

    def as_single_block(self, halo: Optional[PerMember[PerAxis[Halo]]] = None):
        if halo is None:
            halo = {}
        return SampleBlockWithOrigin(
            sample_shape=self.shape,
            sample_id=self.id,
            blocks={
                m: Block(
                    sample_shape=self.shape[m],
                    data=data,
                    inner_slice={
                        a: SliceInfo(0, s) for a, s in data.tagged_shape.items()
                    },
                    halo=halo.get(m, {}),
                    block_index=0,
                    blocks_in_sample=1,
                )
                for m, data in self.members.items()
            },
            stat=self.stat,
            origin=self,
            block_index=0,
            blocks_in_sample=1,
        )

    @classmethod
    def from_blocks(
        cls,
        sample_blocks: Iterable[SampleBlock],
        *,
        fill_value: float = float("nan"),
    ) -> Self:
        members: PerMember[Tensor] = {}
        stat: Stat = {}
        sample_id = None
        for sample_block in sample_blocks:
            assert sample_id is None or sample_id == sample_block.sample_id
            sample_id = sample_block.sample_id
            stat = sample_block.stat
            for m, block in sample_block.blocks.items():
                if m not in members:
                    if -1 in block.sample_shape.values():
                        raise NotImplementedError(
                            "merging blocks with data dependent axis not yet implemented"
                        )

                    members[m] = Tensor(
                        np.full(
                            tuple(block.sample_shape[a] for a in block.data.dims),
                            fill_value,
                            dtype=block.data.dtype,
                        ),
                        dims=block.data.dims,
                    )

                members[m][block.inner_slice] = block.inner_data

        return cls(members=members, stat=stat, id=sample_id)


BlockT = TypeVar("BlockT", Block, BlockMeta)


@dataclass
class SampleBlockBase(Generic[BlockT]):
    """base class for `SampleBlockMeta` and `SampleBlock`"""

    sample_shape: PerMember[PerAxis[int]]
    """the sample shape this block represents a part of"""

    sample_id: SampleId
    """identifier for the sample within its dataset"""

    blocks: Dict[MemberId, BlockT]
    """Individual tensor blocks comprising this sample block"""

    block_index: BlockIndex
    """the n-th block of the sample"""

    blocks_in_sample: TotalNumberOfBlocks
    """total number of blocks in the sample"""

    @property
    def shape(self) -> PerMember[PerAxis[int]]:
        return {mid: b.shape for mid, b in self.blocks.items()}

    @property
    def inner_shape(self) -> PerMember[PerAxis[int]]:
        return {mid: b.inner_shape for mid, b in self.blocks.items()}


@dataclass
class LinearSampleAxisTransform(LinearAxisTransform):
    member: MemberId


@dataclass
class SampleBlockMeta(SampleBlockBase[BlockMeta]):
    """Meta data of a dataset sample block"""

    def get_transformed(
        self, new_axes: PerMember[PerAxis[Union[LinearSampleAxisTransform, int]]]
    ) -> Self:
        sample_shape = {
            m: {
                a: (
                    trf
                    if isinstance(trf, int)
                    else trf.compute(self.sample_shape[trf.member][trf.axis])
                )
                for a, trf in new_axes[m].items()
            }
            for m in new_axes
        }

        def get_member_halo(m: MemberId, round: Callable[[float], int]):
            return {
                a: (
                    Halo(0, 0)
                    if isinstance(trf, int)
                    or trf.axis not in self.blocks[trf.member].halo
                    else Halo(
                        round(self.blocks[trf.member].halo[trf.axis].left * trf.scale),
                        round(self.blocks[trf.member].halo[trf.axis].right * trf.scale),
                    )
                )
                for a, trf in new_axes[m].items()
            }

        halo: Dict[MemberId, Dict[AxisId, Halo]] = {}
        for m in new_axes:
            halo[m] = get_member_halo(m, floor)
            assert halo[m] == get_member_halo(
                m, ceil
            ), f"failed to unambiguously scale halo {halo[m]} with {new_axes[m]}"

        inner_slice = {
            m: {
                a: (
                    SliceInfo(0, trf)
                    if isinstance(trf, int)
                    else SliceInfo(
                        trf.compute(
                            self.blocks[trf.member].inner_slice[trf.axis].start
                        ),
                        trf.compute(self.blocks[trf.member].inner_slice[trf.axis].stop),
                    )
                )
                for a, trf in new_axes[m].items()
            }
            for m in new_axes
        }
        return self.__class__(
            blocks={
                m: BlockMeta(
                    sample_shape=sample_shape[m],
                    inner_slice=inner_slice[m],
                    halo=halo[m],
                    block_index=self.block_index,
                    blocks_in_sample=self.blocks_in_sample,
                )
                for m in new_axes
            },
            sample_shape=sample_shape,
            sample_id=self.sample_id,
            block_index=self.block_index,
            blocks_in_sample=self.blocks_in_sample,
        )

    def with_data(self, data: PerMember[Tensor], *, stat: Stat) -> SampleBlock:
        return SampleBlock(
            sample_shape={
                m: {
                    a: data[m].tagged_shape[a] if s == -1 else s
                    for a, s in member_shape.items()
                }
                for m, member_shape in self.sample_shape.items()
            },
            sample_id=self.sample_id,
            blocks={
                m: Block.from_meta(b, data=data[m]) for m, b in self.blocks.items()
            },
            stat=stat,
            block_index=self.block_index,
            blocks_in_sample=self.blocks_in_sample,
        )


@dataclass
class SampleBlock(SampleBlockBase[Block]):
    """A block of a dataset sample"""

    stat: Stat
    """computed statistics"""

    @property
    def members(self) -> PerMember[Tensor]:
        """the sample block's tensors"""
        return {m: b.data for m, b in self.blocks.items()}

    def get_transformed_meta(
        self, new_axes: PerMember[PerAxis[Union[LinearSampleAxisTransform, int]]]
    ) -> SampleBlockMeta:
        return SampleBlockMeta(
            sample_id=self.sample_id,
            blocks=dict(self.blocks),
            sample_shape=self.sample_shape,
            block_index=self.block_index,
            blocks_in_sample=self.blocks_in_sample,
        ).get_transformed(new_axes)


@dataclass
class SampleBlockWithOrigin(SampleBlock):
    origin: Sample
    """the sample this sample black was taken from"""


class _ConsolidatedMemberBlocks:
    def __init__(self, blocks: PerMember[BlockMeta]):
        super().__init__()
        block_indices = {b.block_index for b in blocks.values()}
        assert len(block_indices) == 1
        self.block_index = block_indices.pop()
        blocks_in_samples = {b.blocks_in_sample for b in blocks.values()}
        assert len(blocks_in_samples) == 1
        self.blocks_in_sample = blocks_in_samples.pop()


def sample_block_meta_generator(
    blocks: Iterable[PerMember[BlockMeta]],
    *,
    sample_shape: PerMember[PerAxis[int]],
    sample_id: SampleId,
):
    for member_blocks in blocks:
        cons = _ConsolidatedMemberBlocks(member_blocks)
        yield SampleBlockMeta(
            blocks=dict(member_blocks),
            sample_shape=sample_shape,
            sample_id=sample_id,
            block_index=cons.block_index,
            blocks_in_sample=cons.blocks_in_sample,
        )


def sample_block_generator(
    blocks: Iterable[PerMember[BlockMeta]],
    *,
    origin: Sample,
    pad_mode: PadMode,
):
    for member_blocks in blocks:
        cons = _ConsolidatedMemberBlocks(member_blocks)
        yield SampleBlockWithOrigin(
            blocks={
                m: Block.from_sample_member(
                    origin.members[m], block=member_blocks[m], pad_mode=pad_mode
                )
                for m in origin.members
            },
            sample_shape=origin.shape,
            origin=origin,
            stat=origin.stat,
            sample_id=origin.id,
            block_index=cons.block_index,
            blocks_in_sample=cons.blocks_in_sample,
        )



# type: ignore  # TODO: type
from pathlib import Path
from typing import List, Sequence, Union

import numpy as np
from numpy.testing import assert_array_almost_equal
from typing_extensions import Any, assert_never

from bioimageio.spec.model import v0_4, v0_5
from bioimageio.spec.model.v0_5 import Version

from ._utils import load_torch_model

try:
    import torch
except ImportError:
    torch = None


# FIXME: remove Any
def _check_predictions(
    model: Any,
    scripted_model: Any,
    model_spec: "v0_4.ModelDescr | v0_5.ModelDescr",
    input_data: Sequence["torch.Tensor"],
):
    assert torch is not None

    def _check(input_: Sequence[torch.Tensor]) -> None:
        expected_tensors = model(*input_)
        if isinstance(expected_tensors, torch.Tensor):
            expected_tensors = [expected_tensors]
        expected_outputs: List[np.ndarray[Any, Any]] = [
            out.numpy() for out in expected_tensors
        ]

        output_tensors = scripted_model(*input_)
        if isinstance(output_tensors, torch.Tensor):
            output_tensors = [output_tensors]
        outputs: List[np.ndarray[Any, Any]] = [out.numpy() for out in output_tensors]

        try:
            for exp, out in zip(expected_outputs, outputs):
                assert_array_almost_equal(exp, out, decimal=4)
        except AssertionError as e:
            raise ValueError(
                f"Results before and after weights conversion do not agree:\n {str(e)}"
            )

    _check(input_data)

    if len(model_spec.inputs) > 1:
        return  # FIXME: why don't we check multiple inputs?

    input_descr = model_spec.inputs[0]
    if isinstance(input_descr, v0_4.InputTensorDescr):
        if not isinstance(input_descr.shape, v0_4.ParameterizedInputShape):
            return
        min_shape = input_descr.shape.min
        step = input_descr.shape.step
    else:
        min_shape: List[int] = []
        step: List[int] = []
        for axis in input_descr.axes:
            if isinstance(axis.size, v0_5.ParameterizedSize):
                min_shape.append(axis.size.min)
                step.append(axis.size.step)
            elif isinstance(axis.size, int):
                min_shape.append(axis.size)
                step.append(0)
            elif axis.size is None:
                raise NotImplementedError(
                    f"Can't verify inputs that don't specify their shape fully: {axis}"
                )
            elif isinstance(axis.size, v0_5.SizeReference):
                raise NotImplementedError(f"Can't handle axes like '{axis}' yet")
            else:
                assert_never(axis.size)

    half_step = [st // 2 for st in step]
    max_steps = 4

    # check that input and output agree for decreasing input sizes
    for step_factor in range(1, max_steps + 1):
        slice_ = tuple(
            slice(None) if st == 0 else slice(step_factor * st, -step_factor * st)
            for st in half_step
        )
        this_input = [inp[slice_] for inp in input_data]
        this_shape = this_input[0].shape
        if any(tsh < msh for tsh, msh in zip(this_shape, min_shape)):
            raise ValueError(
                f"Mismatched shapes: {this_shape}. Expected at least {min_shape}"
            )
        _check(this_input)


def convert_weights_to_torchscript(
    model_descr: Union[v0_4.ModelDescr, v0_5.ModelDescr],
    output_path: Path,
    use_tracing: bool = True,
) -> v0_5.TorchscriptWeightsDescr:
    """Convert model weights from format 'pytorch_state_dict' to 'torchscript'.

    Args:
        model_descr: location of the resource for the input bioimageio model
        output_path: where to save the torchscript weights
        use_tracing: whether to use tracing or scripting to export the torchscript format
    """

    state_dict_weights_descr = model_descr.weights.pytorch_state_dict
    if state_dict_weights_descr is None:
        raise ValueError(
            "The provided model does not have weights in the pytorch state dict format"
        )

    input_data = model_descr.get_input_test_arrays()

    with torch.no_grad():
        input_data = [torch.from_numpy(inp.astype("float32")) for inp in input_data]

        model = load_torch_model(state_dict_weights_descr)

        # FIXME: remove Any
        if use_tracing:
            scripted_model: Any = torch.jit.trace(model, input_data)
        else:
            scripted_model: Any = torch.jit.script(model)

        _check_predictions(
            model=model,
            scripted_model=scripted_model,
            model_spec=model_descr,
            input_data=input_data,
        )

    # save the torchscript model
    scripted_model.save(
        str(output_path)
    )  # does not support Path, so need to cast to str

    return v0_5.TorchscriptWeightsDescr(
        source=output_path,
        pytorch_version=Version(torch.__version__),
        parent="pytorch_state_dict",
    )



# type: ignore  # TODO enable type checking
import os
from pathlib import Path

import pytest


@pytest.mark.skip("onnx converter not updated yet")  # TODO: test onnx converter
def test_onnx_converter(convert_to_onnx: Path, tmp_path: Path):
    from bioimageio.core.weight_converter.torch._onnx import convert_weights_to_onnx

    out_path = tmp_path / "weights.onnx"
    ret_val = convert_weights_to_onnx(convert_to_onnx, out_path, test_decimal=3)
    assert os.path.exists(out_path)
    if not pytest.skip_onnx:
        assert (
            ret_val == 0
        )  # check for correctness is done in converter and returns 0 if it passes



import numpy as np
import pytest
import xarray as xr
from xarray.testing import assert_equal  # pyright: ignore[reportUnknownVariableType]

from bioimageio.core import AxisId, Tensor


@pytest.mark.parametrize(
    "axes",
    ["yx", "xy", "cyx", "yxc", "bczyx", "xyz", "xyzc", "bzyxc"],
)
def test_transpose_tensor_2d(axes: str):

    tensor = Tensor.from_numpy(np.random.rand(256, 256), dims=None)
    transposed = tensor.transpose([AxisId(a) for a in axes])
    assert transposed.ndim == len(axes)


@pytest.mark.parametrize(
    "axes",
    ["zyx", "cyzx", "yzixc", "bczyx", "xyz", "xyzc", "bzyxtc"],
)
def test_transpose_tensor_3d(axes: str):
    tensor = Tensor.from_numpy(np.random.rand(64, 64, 64), dims=None)
    transposed = tensor.transpose([AxisId(a) for a in axes])
    assert transposed.ndim == len(axes)


def test_crop_and_pad():
    tensor = Tensor.from_xarray(
        xr.DataArray(np.random.rand(10, 20), dims=("x", "y"), name="id")
    )
    padded = tensor.pad({AxisId("x"): 7, AxisId("y"): (3, 3)})
    cropped = padded.crop_to(tensor.sizes)
    assert_equal(tensor.data, cropped.data)


def test_some_magic_ops():
    tensor = Tensor.from_numpy(np.random.rand(256, 256), dims=None)
    assert tensor + 2 == 2 + tensor



import pytest

from bioimageio.spec import load_description
from bioimageio.spec.model import v0_5


# TODO: don't just test with unet2d_nuclei_broad_model
@pytest.mark.skip("get_io_sample_block_metas needs improvements")
def test_get_block_transform(unet2d_nuclei_broad_model: str):
    from bioimageio.core.axis import AxisId
    from bioimageio.core.common import MemberId
    from bioimageio.core.digest_spec import (
        get_block_transform,
        get_io_sample_block_metas,
    )

    model = load_description(unet2d_nuclei_broad_model)
    assert isinstance(model, v0_5.ModelDescr)
    block_transform = get_block_transform(model)

    ns = {
        (ipt.id, a.id): 1
        for ipt in model.inputs
        for a in ipt.axes
        if isinstance(a.size, v0_5.ParameterizedSize)
    }

    _, blocks = get_io_sample_block_metas(
        model,
        input_sample_shape={
            MemberId("raw"): {
                AxisId("batch"): 3,
                AxisId("channel"): 1,
                AxisId("x"): 4000,
                AxisId("y"): 3000,
            }
        },
        ns=ns,
    )
    for ipt_block, out_block in blocks:
        trf_block = ipt_block.get_transformed(block_transform)
        assert out_block == trf_block



from typing import Union

from bioimageio.core.model_adapters._pytorch_model_adapter import PytorchModelAdapter
from bioimageio.spec.model import v0_4, v0_5
from bioimageio.spec.utils import download

try:
    import torch
except ImportError:
    torch = None


# additional convenience for pytorch state dict, eventually we want this in python-bioimageio too
# and for each weight format
def load_torch_model(  # pyright: ignore[reportUnknownParameterType]
    node: Union[v0_4.PytorchStateDictWeightsDescr, v0_5.PytorchStateDictWeightsDescr],
):
    assert torch is not None
    model = (  # pyright: ignore[reportUnknownVariableType]
        PytorchModelAdapter.get_network(node)
    )
    state = torch.load(download(node.source).path, map_location="cpu")
    model.load_state_dict(state)  # FIXME: check incompatible keys?
    return model.eval()  # pyright: ignore[reportUnknownVariableType]



import collections.abc
from pathlib import Path
from typing import (
    Any,
    Hashable,
    Iterable,
    Iterator,
    Mapping,
    Optional,
    Tuple,
    Union,
)

import xarray as xr
from loguru import logger
from numpy.typing import NDArray
from tqdm import tqdm

from bioimageio.core.axis import AxisId
from bioimageio.core.io import save_sample
from bioimageio.spec import load_description
from bioimageio.spec.common import PermissiveFileSource
from bioimageio.spec.model import v0_4, v0_5

from ._prediction_pipeline import PredictionPipeline, create_prediction_pipeline
from .common import MemberId, PerMember
from .digest_spec import create_sample_for_model
from .sample import Sample
from .tensor import Tensor


def predict(
    *,
    model: Union[
        PermissiveFileSource, v0_4.ModelDescr, v0_5.ModelDescr, PredictionPipeline
    ],
    inputs: Union[Sample, PerMember[Union[Tensor, xr.DataArray, NDArray[Any], Path]]],
    sample_id: Hashable = "sample",
    blocksize_parameter: Optional[
        Union[
            v0_5.ParameterizedSize_N,
            Mapping[Tuple[MemberId, AxisId], v0_5.ParameterizedSize_N],
        ]
    ] = None,
    input_block_shape: Optional[Mapping[MemberId, Mapping[AxisId, int]]] = None,
    skip_preprocessing: bool = False,
    skip_postprocessing: bool = False,
    save_output_path: Optional[Union[Path, str]] = None,
) -> Sample:
    """Run prediction for a single set of input(s) with a bioimage.io model

    Args:
        model: model to predict with.
            May be given as RDF source, model description or prediction pipeline.
        inputs: the input sample or the named input(s) for this model as a dictionary
        sample_id: the sample id.
        blocksize_parameter: (optional) tile the input into blocks parametrized by
            blocksize according to any parametrized axis sizes defined in the model RDF.
            Note: For a predetermined, fixed block shape use `input_block_shape`
        input_block_shape: (optional) tile the input sample tensors into blocks.
            Note: For a parameterized block shape, not dealing with the exact block shape,
            use `blocksize_parameter`.
        skip_preprocessing: flag to skip the model's preprocessing
        skip_postprocessing: flag to skip the model's postprocessing
        save_output_path: A path with `{member_id}` `{sample_id}` in it
            to save the output to.
    """
    if save_output_path is not None:
        if "{member_id}" not in str(save_output_path):
            raise ValueError(
                f"Missing `{{member_id}}` in save_output_path={save_output_path}"
            )

    if isinstance(model, PredictionPipeline):
        pp = model
    else:
        if not isinstance(model, (v0_4.ModelDescr, v0_5.ModelDescr)):
            loaded = load_description(model)
            if not isinstance(loaded, (v0_4.ModelDescr, v0_5.ModelDescr)):
                raise ValueError(f"expected model description, but got {loaded}")
            model = loaded

        pp = create_prediction_pipeline(model)

    if isinstance(inputs, Sample):
        sample = inputs
    else:
        sample = create_sample_for_model(
            pp.model_description, inputs=inputs, sample_id=sample_id
        )

    if input_block_shape is not None:
        if blocksize_parameter is not None:
            logger.warning(
                "ignoring blocksize_parameter={} in favor of input_block_shape={}",
                blocksize_parameter,
                input_block_shape,
            )

        output = pp.predict_sample_with_fixed_blocking(
            sample,
            input_block_shape=input_block_shape,
            skip_preprocessing=skip_preprocessing,
            skip_postprocessing=skip_postprocessing,
        )
    elif blocksize_parameter is not None:
        output = pp.predict_sample_with_blocking(
            sample,
            skip_preprocessing=skip_preprocessing,
            skip_postprocessing=skip_postprocessing,
            ns=blocksize_parameter,
        )
    else:
        output = pp.predict_sample_without_blocking(
            sample,
            skip_preprocessing=skip_preprocessing,
            skip_postprocessing=skip_postprocessing,
        )
    if save_output_path:
        save_sample(save_output_path, output)

    return output


def predict_many(
    *,
    model: Union[
        PermissiveFileSource, v0_4.ModelDescr, v0_5.ModelDescr, PredictionPipeline
    ],
    inputs: Iterable[PerMember[Union[Tensor, xr.DataArray, NDArray[Any], Path]]],
    sample_id: str = "sample{i:03}",
    blocksize_parameter: Optional[
        Union[
            v0_5.ParameterizedSize_N,
            Mapping[Tuple[MemberId, AxisId], v0_5.ParameterizedSize_N],
        ]
    ] = None,
    skip_preprocessing: bool = False,
    skip_postprocessing: bool = False,
    save_output_path: Optional[Union[Path, str]] = None,
) -> Iterator[Sample]:
    """Run prediction for a multiple sets of inputs with a bioimage.io model

    Args:
        model: model to predict with.
            May be given as RDF source, model description or prediction pipeline.
        inputs: An iterable of the named input(s) for this model as a dictionary.
        sample_id: the sample id.
            note: `{i}` will be formatted as the i-th sample.
            If `{i}` (or `{i:`) is not present and `inputs` is an iterable `{i:03}` is appended.
        blocksize_parameter: (optional) tile the input into blocks parametrized by
            blocksize according to any parametrized axis sizes defined in the model RDF
        skip_preprocessing: flag to skip the model's preprocessing
        skip_postprocessing: flag to skip the model's postprocessing
        save_output_path: A path with `{member_id}` `{sample_id}` in it
            to save the output to.
    """
    if save_output_path is not None:
        if "{member_id}" not in str(save_output_path):
            raise ValueError(
                f"Missing `{{member_id}}` in save_output_path={save_output_path}"
            )

        if not isinstance(inputs, collections.abc.Mapping) and "{sample_id}" not in str(
            save_output_path
        ):
            raise ValueError(
                f"Missing `{{sample_id}}` in save_output_path={save_output_path}"
            )

    if isinstance(model, PredictionPipeline):
        pp = model
    else:
        if not isinstance(model, (v0_4.ModelDescr, v0_5.ModelDescr)):
            loaded = load_description(model)
            if not isinstance(loaded, (v0_4.ModelDescr, v0_5.ModelDescr)):
                raise ValueError(f"expected model description, but got {loaded}")
            model = loaded

        pp = create_prediction_pipeline(model)

    if not isinstance(inputs, collections.abc.Mapping):
        sample_id = str(sample_id)
        if "{i}" not in sample_id and "{i:" not in sample_id:
            sample_id += "{i:03}"

        total = len(inputs) if isinstance(inputs, collections.abc.Sized) else None

        for i, ipts in tqdm(enumerate(inputs), total=total):
            yield predict(
                model=pp,
                inputs=ipts,
                sample_id=sample_id.format(i=i),
                blocksize_parameter=blocksize_parameter,
                skip_preprocessing=skip_preprocessing,
                skip_postprocessing=skip_postprocessing,
                save_output_path=save_output_path,
            )



# TODO: update
# from pathlib import Path

# import imageio
# import numpy as np
# from numpy.testing import assert_array_almost_equal

# from bioimageio.spec import load_description
# from bioimageio.spec.model.v0_4 import InputTensorDescr as InputTensorDescr_v0_4
# from bioimageio.spec.model.v0_4 import ModelDescr as ModelDescr_v0_4
# from bioimageio.spec.model.v0_5 import ModelDescr


# def test_predict_image(any_model: Path, tmpdir: Path):
#     from bioimageio.core.prediction import predict_image

#     spec = load_description(any_model)
#     assert isinstance(spec, ModelDescr)
#     inputs = spec.test_inputs

#     outputs = [Path(tmpdir) / f"out{i}.npy" for i in range(len(spec.test_outputs))]
#     predict_image(any_model, inputs, outputs)
#     for out_path in outputs:
#         assert out_path.exists()

#     result = [np.load(str(p)) for p in outputs]
#     expected = [np.load(str(p)) for p in spec.test_outputs]
#     for res, exp in zip(result, expected):
#         assert_array_almost_equal(res, exp, decimal=4)


# def test_predict_image_with_weight_format(
#     unet2d_fixed_shape_or_not: Path, tmpdir: Path
# ):
#     from bioimageio.core.prediction import predict_image

#     spec = load_description(unet2d_fixed_shape_or_not)
#     assert isinstance(spec, Model)
#     inputs = spec.test_inputs

#     outputs = [Path(tmpdir) / f"out{i}.npy" for i in range(len(spec.test_outputs))]
#     predict_image(
#         unet2d_fixed_shape_or_not, inputs, outputs, weight_format="pytorch_state_dict"
#     )
#     for out_path in outputs:
#         assert out_path.exists()

#     result = [np.load(str(p)) for p in outputs]
#     expected = [np.load(str(p)) for p in spec.test_outputs]
#     for res, exp in zip(result, expected):
#         assert_array_almost_equal(res, exp, decimal=4)


# def _test_predict_with_padding(any_model: Path, tmp_path: Path):
#     from bioimageio.core.digest_spec import get_test_inputs
#     from bioimageio.core.prediction import predict_image

#     model = load_description(any_model)
#     assert isinstance(model, (ModelDescr_v0_4, ModelDescr))

#     input_spec, output_spec = model.inputs[0], model.outputs[0]
#     channel_axis = (
#         "c"
#         if isinstance(input_spec, InputTensorDescr_v0_4)
#         else [a.id for a in input_spec.axes][0]
#     )
#     channel_first = channel_axis == 1

#     # TODO: check more tensors
#     image = get_test_inputs(model)[0]

#     if isinstance(output_spec.shape, list):
#         n_channels = output_spec.shape[channel_axis]
#     else:
#         scale = output_spec.shape.scale[channel_axis]
#         offset = output_spec.shape.offset[channel_axis]
#         in_channels = 1
#         n_channels = int(2 * offset + scale * in_channels)

#     # write the padded image
#     image = image[3:-2, 1:-12]
#     in_path = tmp_path / "in.tif"
#     out_path = tmp_path / "out.tif"
#     imageio.imwrite(in_path, image)

#     if hasattr(output_spec.shape, "scale"):
#         scale = dict(zip(output_spec.axes, output_spec.shape.scale))
#         offset = dict(zip(output_spec.axes, output_spec.shape.offset))
#         spatial_axes = [ax for ax in output_spec.axes if ax in "xyz"]
#         network_resizes = any(
#             sc != 1 for ax, sc in scale.items() if ax in spatial_axes
#         ) or any(off != 0 for ax, off in offset.items() if ax in spatial_axes)
#     else:
#         network_resizes = False

#     if network_resizes:
#         exp_shape = tuple(
#             int(sh * scale[ax] + 2 * offset[ax])
#             for sh, ax in zip(image.shape, spatial_axes)
#         )
#     else:
#         exp_shape = image.shape

#     def check_result():
#         if n_channels == 1:
#             assert out_path.exists()
#             res = imageio.imread(out_path)
#             assert res.shape == exp_shape
#         else:
#             path = str(out_path)
#             for c in range(n_channels):
#                 channel_out_path = Path(path.replace(".tif", f"-c{c}.tif"))
#                 assert channel_out_path.exists()
#                 res = imageio.imread(channel_out_path)
#                 assert res.shape == exp_shape

#     # test with dynamic padding
#     predict_image(
#         any_model, in_path, out_path, padding={"x": 16, "y": 16, "mode": "dynamic"}
#     )
#     check_result()

#     # test with fixed padding
#     predict_image(
#         any_model,
#         in_path,
#         out_path,
#         padding={"x": original_shape[0], "y": original_shape[1], "mode": "fixed"},
#     )
#     check_result()

#     # test with automated padding
#     predict_image(any_model, in_path, out_path, padding=True)
#     check_result()


# # prediction with padding with the parameters above may not be suited for any model
# # so we only run it for the pytorch unet2d here
# def test_predict_image_with_padding(unet2d_fixed_shape_or_not, tmp_path):
#     _test_predict_with_padding(unet2d_fixed_shape_or_not, tmp_path)


# # and with different output shape
# def test_predict_image_with_padding_diff_output_shape(
#     unet2d_diff_output_shape, tmp_path
# ):
#     _test_predict_with_padding(unet2d_diff_output_shape, tmp_path)


# def test_predict_image_with_padding_channel_last(stardist, tmp_path):
#     _test_predict_with_padding(stardist, tmp_path)


# def _test_predict_image_with_tiling(model: Path, tmp_path: Path, exp_mean_deviation):
#     from bioimageio.core.prediction import predict_image

#     spec = load_description(model)
#     assert isinstance(spec, Model)
#     inputs = spec.test_inputs
#     assert len(inputs) == 1
#     exp = np.load(str(spec.test_outputs[0]))

#     out_path = tmp_path.with_suffix(".npy")

#     def check_result():
#         assert out_path.exists()
#         res = np.load(out_path)
#         assert res.shape == exp.shape
#         # check that the mean deviation is smaller than the expected value
#         # note that we can't use array_almost_equal here, because the numerical differences
#         # between tiled and normal prediction are too large
#         mean_deviation = np.abs(res - exp).mean()
#         assert mean_deviation <= exp_mean_deviation

#     # with tiling config
#     tiling = {"halo": {"x": 32, "y": 32}, "tile": {"x": 256, "y": 256}}
#     predict_image(model, inputs, [out_path], tiling=tiling)
#     check_result()

#     # with tiling determined from spec
#     predict_image(model, inputs, [out_path], tiling=True)
#     check_result()


# # prediction with tiling with the parameters above may not be suited for any model
# # so we only run it for the pytorch unet2d here
# def test_predict_image_with_tiling_1(unet2d_nuclei_broad_model: Path, tmp_path: Path):
#     _test_predict_image_with_tiling(unet2d_nuclei_broad_model, tmp_path, 0.012)


# def test_predict_image_with_tiling_2(unet2d_diff_output_shape: Path, tmp_path: Path):
#     _test_predict_image_with_tiling(unet2d_diff_output_shape, tmp_path, 0.06)


# def test_predict_image_with_tiling_3(shape_change_model: Path, tmp_path: Path):
#     _test_predict_image_with_tiling(shape_change_model, tmp_path, 0.012)


# def test_predict_image_with_tiling_channel_last(stardist: Path, tmp_path: Path):
#     _test_predict_image_with_tiling(stardist, tmp_path, 0.13)


# def test_predict_image_with_tiling_fixed_output_shape(
#     unet2d_fixed_shape: Path, tmp_path: Path
# ):
#     _test_predict_image_with_tiling(unet2d_fixed_shape, tmp_path, 0.025)


# def test_predict_images(unet2d_nuclei_broad_model: Path, tmp_path: Path):
#     from bioimageio.core.prediction import predict_images

#     n_images = 5
#     shape = (256, 256)

#     in_paths = []
#     out_paths = []
#     for i in range(n_images):
#         in_path = tmp_path / f"in{i}.tif"
#         im = np.random.randint(0, 255, size=shape).astype("uint8")
#         imageio.imwrite(in_path, im)
#         in_paths.append(in_path)
#         out_paths.append(tmp_path / f"out{i}.tif")
#     predict_images(unet2d_nuclei_broad_model, in_paths, out_paths)

#     for outp in out_paths:
#         assert outp.exists()
#         out = imageio.imread(outp)
#         assert out.shape == shape



from pathlib import Path

from numpy.testing import assert_array_almost_equal

from bioimageio.spec import load_description
from bioimageio.spec.model.v0_4 import ModelDescr as ModelDescr04
from bioimageio.spec.model.v0_5 import ModelDescr, WeightsFormat


def _test_prediction_pipeline(model_package: Path, weights_format: WeightsFormat):
    from bioimageio.core._prediction_pipeline import create_prediction_pipeline
    from bioimageio.core.digest_spec import get_test_inputs, get_test_outputs

    bio_model = load_description(model_package)
    assert isinstance(
        bio_model, (ModelDescr, ModelDescr04)
    ), bio_model.validation_summary.format()
    pp = create_prediction_pipeline(
        bioimageio_model=bio_model, weight_format=weights_format
    )

    inputs = get_test_inputs(bio_model)
    outputs = pp.predict_sample_without_blocking(inputs)

    expected_outputs = get_test_outputs(bio_model)
    assert len(outputs.shape) == len(expected_outputs.shape)
    for m in expected_outputs.members:
        out = outputs.members[m].data
        assert out is not None
        exp = expected_outputs.members[m].data
        assert_array_almost_equal(out, exp, decimal=4)


def test_prediction_pipeline_torch(any_torch_model: Path):
    _test_prediction_pipeline(any_torch_model, "pytorch_state_dict")


def test_prediction_pipeline_torchscript(any_torchscript_model: Path):
    _test_prediction_pipeline(any_torchscript_model, "torchscript")


def test_prediction_pipeline_onnx(any_onnx_model: Path):
    _test_prediction_pipeline(any_onnx_model, "onnx")


def test_prediction_pipeline_tensorflow(any_tensorflow_model: Path):
    _test_prediction_pipeline(any_tensorflow_model, "tensorflow_saved_model_bundle")


def test_prediction_pipeline_keras(any_keras_model: Path):
    _test_prediction_pipeline(any_keras_model, "keras_hdf5")



import zipfile
from typing import List, Literal, Optional, Sequence, Union

import numpy as np
from loguru import logger

from bioimageio.spec.common import FileSource
from bioimageio.spec.model import v0_4, v0_5
from bioimageio.spec.utils import download

from ..digest_spec import get_axes_infos
from ..tensor import Tensor
from ._model_adapter import ModelAdapter

try:
    import tensorflow as tf  # pyright: ignore[reportMissingImports]
except Exception as e:
    tf = None
    tf_error = str(e)
else:
    tf_error = None


class TensorflowModelAdapterBase(ModelAdapter):
    weight_format: Literal["keras_hdf5", "tensorflow_saved_model_bundle"]

    def __init__(
        self,
        *,
        devices: Optional[Sequence[str]] = None,
        weights: Union[
            v0_4.KerasHdf5WeightsDescr,
            v0_4.TensorflowSavedModelBundleWeightsDescr,
            v0_5.KerasHdf5WeightsDescr,
            v0_5.TensorflowSavedModelBundleWeightsDescr,
        ],
        model_description: Union[v0_4.ModelDescr, v0_5.ModelDescr],
    ):
        if tf is None:
            raise ImportError(f"failed to import tensorflow: {tf_error}")

        super().__init__()
        self.model_description = model_description
        tf_version = v0_5.Version(
            tf.__version__  # pyright: ignore[reportUnknownArgumentType]
        )
        model_tf_version = weights.tensorflow_version
        if model_tf_version is None:
            logger.warning(
                "The model does not specify the tensorflow version."
                + f"Cannot check if it is compatible with intalled tensorflow {tf_version}."
            )
        elif model_tf_version > tf_version:
            logger.warning(
                f"The model specifies a newer tensorflow version than installed: {model_tf_version} > {tf_version}."
            )
        elif (model_tf_version.major, model_tf_version.minor) != (
            tf_version.major,
            tf_version.minor,
        ):
            logger.warning(
                "The tensorflow version specified by the model does not match the installed: "
                + f"{model_tf_version} != {tf_version}."
            )

        self.use_keras_api = (
            tf_version.major > 1
            or self.weight_format == KerasModelAdapter.weight_format
        )

        # TODO tf device management
        if devices is not None:
            logger.warning(
                f"Device management is not implemented for tensorflow yet, ignoring the devices {devices}"
            )

        weight_file = self.require_unzipped(weights.source)
        self._network = self._get_network(weight_file)
        self._internal_output_axes = [
            tuple(a.id for a in get_axes_infos(out))
            for out in model_description.outputs
        ]

    def require_unzipped(self, weight_file: FileSource):
        loacl_weights_file = download(weight_file).path
        if zipfile.is_zipfile(loacl_weights_file):
            out_path = loacl_weights_file.with_suffix(".unzipped")
            with zipfile.ZipFile(loacl_weights_file, "r") as f:
                f.extractall(out_path)

            return out_path
        else:
            return loacl_weights_file

    def _get_network(  # pyright: ignore[reportUnknownParameterType]
        self, weight_file: FileSource
    ):
        weight_file = self.require_unzipped(weight_file)
        assert tf is not None
        if self.use_keras_api:
            try:
                return tf.keras.layers.TFSMLayer(
                    weight_file, call_endpoint="serve"
                )  # pyright: ignore[reportUnknownVariableType]
            except Exception as e:
                try:
                    return tf.keras.layers.TFSMLayer(
                        weight_file, call_endpoint="serving_default"
                    )  # pyright: ignore[reportUnknownVariableType]
                except Exception as ee:
                    logger.opt(exception=ee).info(
                        "keras.layers.TFSMLayer error for alternative call_endpoint='serving_default'"
                    )
                    raise e
        else:
            # NOTE in tf1 the model needs to be loaded inside of the session, so we cannot preload the model
            return str(weight_file)

    # TODO currently we relaod the model every time. it would be better to keep the graph and session
    # alive in between of forward passes (but then the sessions need to be properly opened / closed)
    def _forward_tf(  # pyright: ignore[reportUnknownParameterType]
        self, *input_tensors: Optional[Tensor]
    ):
        assert tf is not None
        input_keys = [
            ipt.name if isinstance(ipt, v0_4.InputTensorDescr) else ipt.id
            for ipt in self.model_description.inputs
        ]
        output_keys = [
            out.name if isinstance(out, v0_4.OutputTensorDescr) else out.id
            for out in self.model_description.outputs
        ]
        # TODO read from spec
        tag = (  # pyright: ignore[reportUnknownVariableType]
            tf.saved_model.tag_constants.SERVING
        )
        signature_key = (  # pyright: ignore[reportUnknownVariableType]
            tf.saved_model.signature_constants.DEFAULT_SERVING_SIGNATURE_DEF_KEY
        )

        graph = tf.Graph()  # pyright: ignore[reportUnknownVariableType]
        with graph.as_default():
            with tf.Session(
                graph=graph
            ) as sess:  # pyright: ignore[reportUnknownVariableType]
                # load the model and the signature
                graph_def = tf.saved_model.loader.load(  # pyright: ignore[reportUnknownVariableType]
                    sess, [tag], self._network
                )
                signature = (  # pyright: ignore[reportUnknownVariableType]
                    graph_def.signature_def
                )

                # get the tensors into the graph
                in_names = [  # pyright: ignore[reportUnknownVariableType]
                    signature[signature_key].inputs[key].name for key in input_keys
                ]
                out_names = [  # pyright: ignore[reportUnknownVariableType]
                    signature[signature_key].outputs[key].name for key in output_keys
                ]
                in_tensors = [  # pyright: ignore[reportUnknownVariableType]
                    graph.get_tensor_by_name(name)
                    for name in in_names  # pyright: ignore[reportUnknownVariableType]
                ]
                out_tensors = [  # pyright: ignore[reportUnknownVariableType]
                    graph.get_tensor_by_name(name)
                    for name in out_names  # pyright: ignore[reportUnknownVariableType]
                ]

                # run prediction
                res = sess.run(  # pyright: ignore[reportUnknownVariableType]
                    dict(
                        zip(
                            out_names,  # pyright: ignore[reportUnknownArgumentType]
                            out_tensors,  # pyright: ignore[reportUnknownArgumentType]
                        )
                    ),
                    dict(
                        zip(
                            in_tensors,  # pyright: ignore[reportUnknownArgumentType]
                            input_tensors,
                        )
                    ),
                )
                # from dict to list of tensors
                res = [  # pyright: ignore[reportUnknownVariableType]
                    res[out]
                    for out in out_names  # pyright: ignore[reportUnknownVariableType]
                ]

        return res  # pyright: ignore[reportUnknownVariableType]

    def _forward_keras(  # pyright: ignore[reportUnknownParameterType]
        self, *input_tensors: Optional[Tensor]
    ):
        assert self.use_keras_api
        assert not isinstance(self._network, str)
        assert tf is not None
        tf_tensor = [  # pyright: ignore[reportUnknownVariableType]
            None if ipt is None else tf.convert_to_tensor(ipt) for ipt in input_tensors
        ]

        result = self._network(*tf_tensor)  # pyright: ignore[reportUnknownVariableType]

        assert isinstance(result, dict)

        # TODO: Use RDF's `outputs[i].id` here
        result = list(result.values())

        return [  # pyright: ignore[reportUnknownVariableType]
            (None if r is None else r if isinstance(r, np.ndarray) else r.numpy())
            for r in result  # pyright: ignore[reportUnknownVariableType]
        ]

    def forward(self, *input_tensors: Optional[Tensor]) -> List[Optional[Tensor]]:
        data = [None if ipt is None else ipt.data for ipt in input_tensors]
        if self.use_keras_api:
            result = self._forward_keras(  # pyright: ignore[reportUnknownVariableType]
                *data
            )
        else:
            result = self._forward_tf(  # pyright: ignore[reportUnknownVariableType]
                *data
            )

        return [
            None if r is None else Tensor(r, dims=axes)
            for r, axes in zip(  # pyright: ignore[reportUnknownVariableType]
                result,  # pyright: ignore[reportUnknownArgumentType]
                self._internal_output_axes,
            )
        ]

    def unload(self) -> None:
        logger.warning(
            "Device management is not implemented for keras yet, cannot unload model"
        )


class TensorflowModelAdapter(TensorflowModelAdapterBase):
    weight_format = "tensorflow_saved_model_bundle"

    def __init__(
        self,
        *,
        model_description: Union[v0_4.ModelDescr, v0_5.ModelDescr],
        devices: Optional[Sequence[str]] = None,
    ):
        if model_description.weights.tensorflow_saved_model_bundle is None:
            raise ValueError("missing tensorflow_saved_model_bundle weights")

        super().__init__(
            devices=devices,
            weights=model_description.weights.tensorflow_saved_model_bundle,
            model_description=model_description,
        )


class KerasModelAdapter(TensorflowModelAdapterBase):
    weight_format = "keras_hdf5"

    def __init__(
        self,
        *,
        model_description: Union[v0_4.ModelDescr, v0_5.ModelDescr],
        devices: Optional[Sequence[str]] = None,
    ):
        if model_description.weights.keras_hdf5 is None:
            raise ValueError("missing keras_hdf5 weights")

        super().__init__(
            model_description=model_description,
            devices=devices,
            weights=model_description.weights.keras_hdf5,
        )



from pathlib import Path

from numpy.testing import assert_array_almost_equal

from bioimageio.core.utils.testing import skip_on
from bioimageio.spec.model.v0_4 import ModelDescr as ModelDescr04
from bioimageio.spec.model.v0_5 import ModelDescr, WeightsFormat


class TooFewDevicesException(Exception):
    pass


def _test_device_management(model_package: Path, weight_format: WeightsFormat):
    import torch

    from bioimageio.core import load_description
    from bioimageio.core._prediction_pipeline import create_prediction_pipeline
    from bioimageio.core.digest_spec import get_test_inputs, get_test_outputs

    if not hasattr(torch, "cuda") or torch.cuda.device_count() == 0:
        raise TooFewDevicesException("Need at least one cuda device for this test")

    bio_model = load_description(model_package)
    assert isinstance(bio_model, (ModelDescr, ModelDescr04))
    pred_pipe = create_prediction_pipeline(
        bioimageio_model=bio_model, weight_format=weight_format, devices=["cuda:0"]
    )

    inputs = get_test_inputs(bio_model)
    with pred_pipe as pp:
        outputs = pp.predict_sample_without_blocking(inputs)

    expected_outputs = get_test_outputs(bio_model)

    assert len(outputs.shape) == len(expected_outputs.shape)
    for m in expected_outputs.members:
        out = outputs.members[m].data
        assert out is not None
        exp = expected_outputs.members[m].data
        assert_array_almost_equal(out, exp, decimal=4)

    # repeat inference with context manager to test load/predict/unload/load/predict
    with pred_pipe as pp:
        outputs = pp.predict_sample_without_blocking(inputs)

    assert len(outputs.shape) == len(expected_outputs.shape)
    for m in expected_outputs.members:
        out = outputs.members[m].data
        assert out is not None
        exp = expected_outputs.members[m].data
        assert_array_almost_equal(out, exp, decimal=4)


@skip_on(
    TooFewDevicesException, reason="Too few devices"
)  # pyright: ignore[reportArgumentType]
def test_device_management_torch(any_torch_model: Path):
    _test_device_management(any_torch_model, "pytorch_state_dict")


@skip_on(
    TooFewDevicesException, reason="Too few devices"
)  # pyright: ignore[reportArgumentType]
def test_device_management_torchscript(any_torchscript_model: Path):
    _test_device_management(any_torchscript_model, "torchscript")


@skip_on(
    TooFewDevicesException, reason="Too few devices"
)  # pyright: ignore[reportArgumentType]
def test_device_management_onnx(any_onnx_model: Path):
    _test_device_management(any_onnx_model, "onnx")


@skip_on(
    TooFewDevicesException, reason="Too few devices"
)  # pyright: ignore[reportArgumentType]
def test_device_management_tensorflow(any_tensorflow_model: Path):
    _test_device_management(any_tensorflow_model, "tensorflow_saved_model_bundle")


@skip_on(
    TooFewDevicesException, reason="Too few devices"
)  # pyright: ignore[reportArgumentType]
def test_device_management_keras(any_keras_model: Path):
    _test_device_management(any_keras_model, "keras_hdf5")



import gc
import warnings
from typing import Any, List, Optional, Sequence, Tuple, Union

from bioimageio.spec.model import v0_4, v0_5
from bioimageio.spec.utils import download

from ..axis import AxisId
from ..digest_spec import get_axes_infos, import_callable
from ..tensor import Tensor
from ._model_adapter import ModelAdapter

try:
    import torch
except Exception as e:
    torch = None
    torch_error = str(e)
else:
    torch_error = None


class PytorchModelAdapter(ModelAdapter):
    def __init__(
        self,
        *,
        outputs: Union[
            Sequence[v0_4.OutputTensorDescr], Sequence[v0_5.OutputTensorDescr]
        ],
        weights: Union[
            v0_4.PytorchStateDictWeightsDescr, v0_5.PytorchStateDictWeightsDescr
        ],
        devices: Optional[Sequence[str]] = None,
    ):
        if torch is None:
            raise ImportError(f"failed to import torch: {torch_error}")

        super().__init__()
        self.output_dims = [tuple(a.id for a in get_axes_infos(out)) for out in outputs]
        self._network = self.get_network(weights)
        self._devices = self.get_devices(devices)
        self._network = self._network.to(self._devices[0])

        self._primary_device = self._devices[0]
        state: Any = torch.load(
            download(weights).path,
            map_location=self._primary_device,  # pyright: ignore[reportUnknownArgumentType]
        )
        self._network.load_state_dict(state)

        self._network = self._network.eval()

    def forward(self, *input_tensors: Optional[Tensor]) -> List[Optional[Tensor]]:
        if torch is None:
            raise ImportError("torch")
        with torch.no_grad():
            tensors = [
                None if ipt is None else torch.from_numpy(ipt.data.data)
                for ipt in input_tensors
            ]
            tensors = [
                (
                    None
                    if t is None
                    else t.to(
                        self._primary_device  # pyright: ignore[reportUnknownArgumentType]
                    )
                )
                for t in tensors
            ]
            result: Union[Tuple[Any, ...], List[Any], Any]
            result = self._network(  # pyright: ignore[reportUnknownVariableType]
                *tensors
            )
            if not isinstance(result, (tuple, list)):
                result = [result]

            result = [
                (
                    None
                    if r is None
                    else r.detach().cpu().numpy() if isinstance(r, torch.Tensor) else r
                )
                for r in result  # pyright: ignore[reportUnknownVariableType]
            ]
            if len(result) > len(self.output_dims):
                raise ValueError(
                    f"Expected at most {len(self.output_dims)} outputs, but got {len(result)}"
                )

        return [
            None if r is None else Tensor(r, dims=out)
            for r, out in zip(result, self.output_dims)
        ]

    def unload(self) -> None:
        del self._network
        _ = gc.collect()  # deallocate memory
        assert torch is not None
        torch.cuda.empty_cache()  # release reserved memory

    @staticmethod
    def get_network(  # pyright: ignore[reportUnknownParameterType]
        weight_spec: Union[
            v0_4.PytorchStateDictWeightsDescr, v0_5.PytorchStateDictWeightsDescr
        ]
    ) -> "torch.nn.Module":  # pyright: ignore[reportInvalidTypeForm]
        if torch is None:
            raise ImportError("torch")
        arch = import_callable(
            weight_spec.architecture,
            sha256=(
                weight_spec.architecture_sha256
                if isinstance(weight_spec, v0_4.PytorchStateDictWeightsDescr)
                else weight_spec.sha256
            ),
        )
        model_kwargs = (
            weight_spec.kwargs
            if isinstance(weight_spec, v0_4.PytorchStateDictWeightsDescr)
            else weight_spec.architecture.kwargs
        )
        network = arch(**model_kwargs)
        if not isinstance(network, torch.nn.Module):
            raise ValueError(
                f"calling {weight_spec.architecture.callable} did not return a torch.nn.Module"
            )

        return network

    @staticmethod
    def get_devices(  # pyright: ignore[reportUnknownParameterType]
        devices: Optional[Sequence[str]] = None,
    ) -> List["torch.device"]:  # pyright: ignore[reportInvalidTypeForm]
        if torch is None:
            raise ImportError("torch")
        if not devices:
            torch_devices = [
                (
                    torch.device("cuda")
                    if torch.cuda.is_available()
                    else torch.device("cpu")
                )
            ]
        else:
            torch_devices = [torch.device(d) for d in devices]

        if len(torch_devices) > 1:
            warnings.warn(
                f"Multiple devices for single pytorch model not yet implemented; ignoring {torch_devices[1:]}"
            )
            torch_devices = torch_devices[:1]

        return torch_devices



"""
.. include:: ../../README.md
"""

from bioimageio.spec import build_description as build_description
from bioimageio.spec import dump_description as dump_description
from bioimageio.spec import load_dataset_description as load_dataset_description
from bioimageio.spec import load_description as load_description
from bioimageio.spec import (
    load_description_and_validate_format_only as load_description_and_validate_format_only,
)
from bioimageio.spec import load_model_description as load_model_description
from bioimageio.spec import save_bioimageio_package as save_bioimageio_package
from bioimageio.spec import (
    save_bioimageio_package_as_folder as save_bioimageio_package_as_folder,
)
from bioimageio.spec import save_bioimageio_yaml_only as save_bioimageio_yaml_only
from bioimageio.spec import validate_format as validate_format

from . import digest_spec as digest_spec
from ._prediction_pipeline import PredictionPipeline as PredictionPipeline
from ._prediction_pipeline import (
    create_prediction_pipeline as create_prediction_pipeline,
)
from ._resource_tests import load_description_and_test as load_description_and_test
from ._resource_tests import test_description as test_description
from ._resource_tests import test_model as test_model
from ._settings import settings as settings
from .axis import Axis as Axis
from .axis import AxisId as AxisId
from .block_meta import BlockMeta as BlockMeta
from .common import MemberId as MemberId
from .prediction import predict as predict
from .prediction import predict_many as predict_many
from .sample import Sample as Sample
from .tensor import Tensor as Tensor
from .utils import VERSION

__version__ = VERSION

# aliases
test_resource = test_description
load_resource = load_description
load_model = load_model_description



import json
from pathlib import Path

from setuptools import find_namespace_packages, setup

# Get the long description from the README file
ROOT_DIR = Path(__file__).parent.resolve()
long_description = (ROOT_DIR / "README.md").read_text(encoding="utf-8")
VERSION_FILE = ROOT_DIR / "bioimageio" / "core" / "VERSION"
VERSION = json.loads(VERSION_FILE.read_text())["version"]


_ = setup(
    name="bioimageio.core",
    version=VERSION,
    description="Python functionality for the bioimage model zoo",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/bioimage-io/core-bioimage-io-python",
    author="Bioimage Team",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
    ],
    packages=find_namespace_packages(exclude=["tests"]),
    install_requires=[
        "bioimageio.spec ==0.5.3.3",
        "imageio>=2.10",
        "loguru",
        "numpy",
        "pydantic-settings >=2.3",
        "pydantic",
        "python-dotenv",
        "requests",
        "ruyaml",
        "tqdm",
        "typing-extensions",
        "xarray",
    ],
    include_package_data=True,
    extras_require={
        "pytorch": ["torch>=1.6", "torchvision", "keras>=3.0"],
        "tensorflow": ["tensorflow", "keras>=2.15"],
        "onnx": ["onnxruntime"],
        "dev": [
            "black",
            # "crick",  # currently requires python<=3.9
            "filelock",
            "jupyter",
            "jupyter-black",
            "matplotlib",
            "keras>=3.0",
            "onnxruntime",
            "packaging>=17.0",
            "pre-commit",
            "pdoc",
            "psutil",  # parallel pytest with 'pytest -n auto'
            "pyright",
            "pytest-xdist",  # parallel pytest
            "pytest",
            "torch>=1.6",
            "torchvision",
        ],
    },
    project_urls={
        "Bug Reports": "https://github.com/bioimage-io/core-bioimage-io-python/issues",
        "Source": "https://github.com/bioimage-io/core-bioimage-io-python",
    },
    entry_points={"console_scripts": ["bioimageio = bioimageio.core.__main__:main"]},
)



# type: ignore  # TODO: type
import warnings
from pathlib import Path
from typing import Any, List, Sequence, cast

import numpy as np
from numpy.testing import assert_array_almost_equal

from bioimageio.spec import load_description
from bioimageio.spec.common import InvalidDescr
from bioimageio.spec.model import v0_4, v0_5

from ...digest_spec import get_member_id, get_test_inputs
from ...weight_converter.torch._utils import load_torch_model

try:
    import torch
except ImportError:
    torch = None


def add_onnx_weights(
    model_spec: "str | Path | v0_4.ModelDescr | v0_5.ModelDescr",
    *,
    output_path: Path,
    use_tracing: bool = True,
    test_decimal: int = 4,
    verbose: bool = False,
    opset_version: "int | None" = None,
):
    """Convert model weights from format 'pytorch_state_dict' to 'onnx'.

    Args:
        source_model: model without onnx weights
        opset_version: onnx opset version
        use_tracing: whether to use tracing or scripting to export the onnx format
        test_decimal: precision for testing whether the results agree
    """
    if isinstance(model_spec, (str, Path)):
        loaded_spec = load_description(Path(model_spec))
        if isinstance(loaded_spec, InvalidDescr):
            raise ValueError(f"Bad resource description: {loaded_spec}")
        if not isinstance(loaded_spec, (v0_4.ModelDescr, v0_5.ModelDescr)):
            raise TypeError(
                f"Path {model_spec} is a {loaded_spec.__class__.__name__}, expected a v0_4.ModelDescr or v0_5.ModelDescr"
            )
        model_spec = loaded_spec

    state_dict_weights_descr = model_spec.weights.pytorch_state_dict
    if state_dict_weights_descr is None:
        raise ValueError(
            "The provided model does not have weights in the pytorch state dict format"
        )

    assert torch is not None
    with torch.no_grad():

        sample = get_test_inputs(model_spec)
        input_data = [sample[get_member_id(ipt)].data.data for ipt in model_spec.inputs]
        input_tensors = [torch.from_numpy(ipt) for ipt in input_data]
        model = load_torch_model(state_dict_weights_descr)

        expected_tensors = model(*input_tensors)
        if isinstance(expected_tensors, torch.Tensor):
            expected_tensors = [expected_tensors]
        expected_outputs: List[np.ndarray[Any, Any]] = [
            out.numpy() for out in expected_tensors
        ]

        if use_tracing:
            torch.onnx.export(
                model,
                tuple(input_tensors) if len(input_tensors) > 1 else input_tensors[0],
                str(output_path),
                verbose=verbose,
                opset_version=opset_version,
            )
        else:
            raise NotImplementedError

    try:
        import onnxruntime as rt  # pyright: ignore [reportMissingTypeStubs]
    except ImportError:
        msg = "The onnx weights were exported, but onnx rt is not available and weights cannot be checked."
        warnings.warn(msg)
        return

    # check the onnx model
    sess = rt.InferenceSession(str(output_path))
    onnx_input_node_args = cast(
        List[Any], sess.get_inputs()
    )  # fixme: remove cast, try using rt.NodeArg instead of Any
    onnx_inputs = {
        input_name.name: inp
        for input_name, inp in zip(onnx_input_node_args, input_data)
    }
    outputs = cast(
        Sequence[np.ndarray[Any, Any]], sess.run(None, onnx_inputs)
    )  # FIXME: remove cast

    try:
        for exp, out in zip(expected_outputs, outputs):
            assert_array_almost_equal(exp, out, decimal=test_decimal)
        return 0
    except AssertionError as e:
        msg = f"The onnx weights were exported, but results before and after conversion do not agree:\n {str(e)}"
        warnings.warn(msg)
        return 1



# TODO: move to tests/
from functools import wraps
from typing import Any, Protocol, Type


class test_func(Protocol):
    def __call__(*args: Any, **kwargs: Any): ...


def skip_on(exception: Type[Exception], reason: str):
    """adapted from https://stackoverflow.com/a/63522579"""
    import pytest

    # Func below is the real decorator and will receive the test function as param
    def decorator_func(f: test_func):
        @wraps(f)
        def wrapper(*args: Any, **kwargs: Any):
            try:
                # Try to run the test
                return f(*args, **kwargs)
            except exception:
                # If exception of given type happens
                # just swallow it and raise pytest.Skip with given reason
                pytest.skip(reason)

        return wrapper

    return decorator_func



from typing import Literal

from dotenv import load_dotenv
from pydantic import Field
from typing_extensions import Annotated

from bioimageio.spec._internal._settings import Settings as SpecSettings

_ = load_dotenv()


class Settings(SpecSettings):
    """environment variables"""

    keras_backend: Annotated[
        Literal["torch", "tensorflow", "jax"], Field(alias="KERAS_BACKEND")
    ] = "torch"


settings = Settings()



"""These functions implement the logic of the bioimageio command line interface
defined in the `cli` module."""

import sys
from pathlib import Path
from typing import List, Optional, Sequence, Union

from typing_extensions import Literal

from bioimageio.core import test_description
from bioimageio.spec import (
    InvalidDescr,
    ResourceDescr,
    save_bioimageio_package,
    save_bioimageio_package_as_folder,
)
from bioimageio.spec.model.v0_5 import WeightsFormat

WeightFormatArgAll = Literal[WeightsFormat, "all"]
WeightFormatArgAny = Literal[WeightsFormat, "any"]


def test(
    descr: Union[ResourceDescr, InvalidDescr],
    *,
    weight_format: WeightFormatArgAll = "all",
    devices: Optional[Union[str, Sequence[str]]] = None,
    decimal: int = 4,
):
    """test a bioimageio resource

    Args:
        source: Path or URL to the bioimageio resource description file
                (bioimageio.yaml or rdf.yaml) or to a zipped resource
        weight_format: (model only) The weight format to use
        devices: Device(s) to use for testing
        decimal: Precision for numerical comparisons
    """
    if isinstance(descr, InvalidDescr):
        descr.validation_summary.display()
        sys.exit(1)

    summary = test_description(
        descr,
        weight_format=None if weight_format == "all" else weight_format,
        devices=[devices] if isinstance(devices, str) else devices,
        decimal=decimal,
    )
    summary.display()
    sys.exit(0 if summary.status == "passed" else 1)


def validate_format(
    descr: Union[ResourceDescr, InvalidDescr],
):
    """validate the meta data format of a bioimageio resource

    Args:
        descr: a bioimageio resource description
    """
    descr.validation_summary.display()
    sys.exit(0 if descr.validation_summary.status == "passed" else 1)


def package(
    descr: ResourceDescr, path: Path, *, weight_format: WeightFormatArgAll = "all"
):
    """Save a resource's metadata with its associated files.

    Note: If `path` does not have a `.zip` suffix this command will save the
          package as an unzipped folder instead.

    Args:
        descr: a bioimageio resource description
        path: output path
        weight-format: include only this single weight-format (if not 'all').
    """
    if isinstance(descr, InvalidDescr):
        descr.validation_summary.display()
        raise ValueError("resource description is invalid")

    if weight_format == "all":
        weights_priority_order = None
    else:
        weights_priority_order = (weight_format,)

    if path.suffix == ".zip":
        _ = save_bioimageio_package(
            descr,
            output_path=path,
            weights_priority_order=weights_priority_order,
        )
    else:
        _ = save_bioimageio_package_as_folder(
            descr,
            output_path=path,
            weights_priority_order=weights_priority_order,
        )


# TODO: add convert command(s)
# if torch_converter is not None:

#     @app.command()
#     def convert_torch_weights_to_onnx(
#         model_rdf: Path = typer.Argument(
#             ..., help="Path to the model resource description file (rdf.yaml) or zipped model."
#         ),
#         output_path: Path = typer.Argument(..., help="Where to save the onnx weights."),
#         opset_version: Optional[int] = typer.Argument(12, help="Onnx opset version."),
#         use_tracing: bool = typer.Option(True, help="Whether to use torch.jit tracing or scripting."),
#         verbose: bool = typer.Option(True, help="Verbosity"),
#     ):
#         ret_code = torch_converter.convert_weights_to_onnx(model_rdf, output_path, opset_version, use_tracing, verbose)
#         sys.exit(ret_code)

#     convert_torch_weights_to_onnx.__doc__ = torch_converter.convert_weights_to_onnx.__doc__

#     @app.command()
#     def convert_torch_weights_to_torchscript(
#         model_rdf: Path = typer.Argument(
#             ..., help="Path to the model resource description file (rdf.yaml) or zipped model."
#         ),
#         output_path: Path = typer.Argument(..., help="Where to save the torchscript weights."),
#         use_tracing: bool = typer.Option(True, help="Whether to use torch.jit tracing or scripting."),
#     ):
#         torch_converter.convert_weights_to_torchscript(model_rdf, output_path, use_tracing)
#         sys.exit(0)

#     convert_torch_weights_to_torchscript.__doc__ = torch_converter.convert_weights_to_torchscript.__doc__


# if keras_converter is not None:

#     @app.command()
#     def convert_keras_weights_to_tensorflow(
#         model_rdf: Annotated[
#             Path, typer.Argument(help="Path to the model resource description file (rdf.yaml) or zipped model.")
#         ],
#         output_path: Annotated[Path, typer.Argument(help="Where to save the tensorflow weights.")],
#     ):
#         rd = load_description(model_rdf)
#         ret_code = keras_converter.convert_weights_to_tensorflow_saved_model_bundle(rd, output_path)
#         sys.exit(ret_code)

#     convert_keras_weights_to_tensorflow.__doc__ = (
#         keras_converter.convert_weights_to_tensorflow_saved_model_bundle.__doc__
#     )



import os
from typing import Any, List, Optional, Sequence, Union

from loguru import logger
from numpy.typing import NDArray

from bioimageio.spec._internal.io_utils import download
from bioimageio.spec.model import v0_4, v0_5
from bioimageio.spec.model.v0_5 import Version

from .._settings import settings
from ..digest_spec import get_axes_infos
from ..tensor import Tensor
from ._model_adapter import ModelAdapter

os.environ["KERAS_BACKEND"] = settings.keras_backend

# by default, we use the keras integrated with tensorflow
try:
    import tensorflow as tf  # pyright: ignore[reportMissingImports]
    from tensorflow import (  # pyright: ignore[reportMissingImports]
        keras,  # pyright: ignore[reportUnknownVariableType]
    )

    tf_version = Version(tf.__version__)  # pyright: ignore[reportUnknownArgumentType]
except Exception:
    try:
        import keras  # pyright: ignore[reportMissingImports]
    except Exception as e:
        keras = None
        keras_error = str(e)
    else:
        keras_error = None
    tf_version = None
else:
    keras_error = None


class KerasModelAdapter(ModelAdapter):
    def __init__(
        self,
        *,
        model_description: Union[v0_4.ModelDescr, v0_5.ModelDescr],
        devices: Optional[Sequence[str]] = None,
    ) -> None:
        if keras is None:
            raise ImportError(f"failed to import keras: {keras_error}")

        super().__init__()
        if model_description.weights.keras_hdf5 is None:
            raise ValueError("model has not keras_hdf5 weights specified")
        model_tf_version = model_description.weights.keras_hdf5.tensorflow_version

        if tf_version is None or model_tf_version is None:
            logger.warning("Could not check tensorflow versions.")
        elif model_tf_version > tf_version:
            logger.warning(
                "The model specifies a newer tensorflow version than installed: {} > {}.",
                model_tf_version,
                tf_version,
            )
        elif (model_tf_version.major, model_tf_version.minor) != (
            tf_version.major,
            tf_version.minor,
        ):
            logger.warning(
                "Model tensorflow version {} does not match {}.",
                model_tf_version,
                tf_version,
            )

        # TODO keras device management
        if devices is not None:
            logger.warning(
                "Device management is not implemented for keras yet, ignoring the devices {}",
                devices,
            )

        weight_path = download(model_description.weights.keras_hdf5.source).path

        self._network = keras.models.load_model(weight_path)
        self._output_axes = [
            tuple(a.id for a in get_axes_infos(out))
            for out in model_description.outputs
        ]

    def forward(self, *input_tensors: Optional[Tensor]) -> List[Optional[Tensor]]:
        _result: Union[Sequence[NDArray[Any]], NDArray[Any]]
        _result = self._network.predict(  # pyright: ignore[reportUnknownVariableType]
            *[None if t is None else t.data.data for t in input_tensors]
        )
        if isinstance(_result, (tuple, list)):
            result: Sequence[NDArray[Any]] = _result
        else:
            result = [_result]  # type: ignore

        assert len(result) == len(self._output_axes)
        ret: List[Optional[Tensor]] = []
        ret.extend(
            [Tensor(r, dims=axes) for r, axes, in zip(result, self._output_axes)]
        )
        return ret

    def unload(self) -> None:
        logger.warning(
            "Device management is not implemented for keras yet, cannot unload model"
        )



import json
import subprocess
from typing import Optional

import pytest
from packaging.version import Version


def test_bioimageio_spec_version(mamba_cmd: Optional[str]):
    if mamba_cmd is None:
        pytest.skip("requires mamba")

    from importlib.metadata import metadata

    # get latest released bioimageio.spec version
    mamba_repoquery = subprocess.run(
        f"{mamba_cmd} repoquery search -c conda-forge --json bioimageio.spec".split(
            " "
        ),
        encoding="utf-8",
        capture_output=True,
        check=True,
    )
    full_out = mamba_repoquery.stdout  # full output includes mamba banner
    search = json.loads(full_out[full_out.find("{") :])  # json output starts at '{'
    latest_spec = max(search["result"]["pkgs"], key=lambda entry: entry["timestamp"])
    rmaj, rmin, rpatch, *_ = latest_spec["version"].split(".")
    released = Version(f"{rmaj}.{rmin}.{rpatch}")

    # get currently pinned bioimageio.spec version
    meta = metadata("bioimageio.core")
    req = meta["Requires-Dist"]
    valid_starts = ("bioimageio.spec ==", "bioimageio.spec==")
    for start in valid_starts:
        if req.startswith(start):
            spec_ver = req[len(start) :]
            break
    else:
        raise ValueError(
            f"Expected bioimageio.sepc pin to start with any of {valid_starts}"
        )

    assert spec_ver.count(".") == 3
    pmaj, pmin, ppatch, _ = spec_ver.split(".")
    assert (
        pmaj.isdigit() and pmin.isdigit() and ppatch.isdigit()
    ), "bioimageio.spec version should be pinned down to patch, e.g. '0.4.9.*'"

    pinned = Version(f"{pmaj}.{pmin}.{ppatch}")
    assert pinned == released, "bioimageio.spec not pinned to the latest version"



import traceback
import warnings
from itertools import product
from typing import Dict, Hashable, List, Literal, Optional, Sequence, Set, Tuple, Union

import numpy as np
from loguru import logger

from bioimageio.core.sample import Sample
from bioimageio.spec import (
    InvalidDescr,
    ResourceDescr,
    build_description,
    dump_description,
    load_description,
)
from bioimageio.spec._internal.common_nodes import ResourceDescrBase
from bioimageio.spec.common import BioimageioYamlContent, PermissiveFileSource
from bioimageio.spec.model import v0_4, v0_5
from bioimageio.spec.model.v0_5 import WeightsFormat
from bioimageio.spec.summary import (
    ErrorEntry,
    InstalledPackage,
    ValidationDetail,
    ValidationSummary,
)

from ._prediction_pipeline import create_prediction_pipeline
from .axis import AxisId, BatchSize
from .digest_spec import get_test_inputs, get_test_outputs
from .utils import VERSION


def test_model(
    source: Union[v0_5.ModelDescr, PermissiveFileSource],
    weight_format: Optional[WeightsFormat] = None,
    devices: Optional[List[str]] = None,
    absolute_tolerance: float = 1.5e-4,
    relative_tolerance: float = 1e-4,
    decimal: Optional[int] = None,
) -> ValidationSummary:
    """Test model inference"""
    # NOTE: `decimal` is a legacy argument and is handled in `_test_model_inference`
    return test_description(
        source,
        weight_format=weight_format,
        devices=devices,
        absolute_tolerance=absolute_tolerance,
        relative_tolerance=relative_tolerance,
        decimal=decimal,
        expected_type="model",
    )


def test_description(
    source: Union[ResourceDescr, PermissiveFileSource, BioimageioYamlContent],
    *,
    format_version: Union[Literal["discover", "latest"], str] = "discover",
    weight_format: Optional[WeightsFormat] = None,
    devices: Optional[Sequence[str]] = None,
    absolute_tolerance: float = 1.5e-4,
    relative_tolerance: float = 1e-4,
    decimal: Optional[int] = None,
    expected_type: Optional[str] = None,
) -> ValidationSummary:
    """Test a bioimage.io resource dynamically, e.g. prediction of test tensors for models"""
    # NOTE: `decimal` is a legacy argument and is handled in `_test_model_inference`
    rd = load_description_and_test(
        source,
        format_version=format_version,
        weight_format=weight_format,
        devices=devices,
        absolute_tolerance=absolute_tolerance,
        relative_tolerance=relative_tolerance,
        decimal=decimal,
        expected_type=expected_type,
    )
    return rd.validation_summary


def load_description_and_test(
    source: Union[ResourceDescr, PermissiveFileSource, BioimageioYamlContent],
    *,
    format_version: Union[Literal["discover", "latest"], str] = "discover",
    weight_format: Optional[WeightsFormat] = None,
    devices: Optional[Sequence[str]] = None,
    absolute_tolerance: float = 1.5e-4,
    relative_tolerance: float = 1e-4,
    decimal: Optional[int] = None,
    expected_type: Optional[str] = None,
) -> Union[ResourceDescr, InvalidDescr]:
    """Test RDF dynamically, e.g. model inference of test inputs"""
    # NOTE: `decimal` is a legacy argument and is handled in `_test_model_inference`
    if (
        isinstance(source, ResourceDescrBase)
        and format_version != "discover"
        and source.format_version != format_version
    ):
        warnings.warn(
            f"deserializing source to ensure we validate and test using format {format_version}"
        )
        source = dump_description(source)

    if isinstance(source, ResourceDescrBase):
        rd = source
    elif isinstance(source, dict):
        rd = build_description(source, format_version=format_version)
    else:
        rd = load_description(source, format_version=format_version)

    rd.validation_summary.env.append(
        InstalledPackage(name="bioimageio.core", version=VERSION)
    )

    if expected_type is not None:
        _test_expected_resource_type(rd, expected_type)

    if isinstance(rd, (v0_4.ModelDescr, v0_5.ModelDescr)):
        if weight_format is None:
            weight_formats: List[WeightsFormat] = [
                w for w, we in rd.weights if we is not None
            ]  # pyright: ignore[reportAssignmentType]
        else:
            weight_formats = [weight_format]
        for w in weight_formats:
            _test_model_inference(
                rd, w, devices, absolute_tolerance, relative_tolerance, decimal
            )
            if not isinstance(rd, v0_4.ModelDescr):
                _test_model_inference_parametrized(rd, w, devices)

    # TODO: add execution of jupyter notebooks
    # TODO: add more tests

    return rd


def _test_model_inference(
    model: Union[v0_4.ModelDescr, v0_5.ModelDescr],
    weight_format: WeightsFormat,
    devices: Optional[Sequence[str]],
    absolute_tolerance: float,
    relative_tolerance: float,
    decimal: Optional[int],
) -> None:
    test_name = f"Reproduce test outputs from test inputs ({weight_format})"
    logger.info("starting '{}'", test_name)
    error: Optional[str] = None
    tb: List[str] = []

    precision_args = _handle_legacy_precision_args(
        absolute_tolerance=absolute_tolerance,
        relative_tolerance=relative_tolerance,
        decimal=decimal,
    )

    try:
        inputs = get_test_inputs(model)
        expected = get_test_outputs(model)

        with create_prediction_pipeline(
            bioimageio_model=model, devices=devices, weight_format=weight_format
        ) as prediction_pipeline:
            results = prediction_pipeline.predict_sample_without_blocking(inputs)

        if len(results.members) != len(expected.members):
            error = f"Expected {len(expected.members)} outputs, but got {len(results.members)}"

        else:
            for m, exp in expected.members.items():
                res = results.members.get(m)
                if res is None:
                    error = "Output tensors for test case may not be None"
                    break
                try:
                    np.testing.assert_allclose(
                        res.data,
                        exp.data,
                        rtol=precision_args["relative_tolerance"],
                        atol=precision_args["absolute_tolerance"],
                    )
                except AssertionError as e:
                    error = f"Output and expected output disagree:\n {e}"
                    break
    except Exception as e:
        error = str(e)
        tb = traceback.format_tb(e.__traceback__)

    model.validation_summary.add_detail(
        ValidationDetail(
            name=test_name,
            status="passed" if error is None else "failed",
            errors=(
                []
                if error is None
                else [
                    ErrorEntry(
                        loc=("weights", weight_format),
                        msg=error,
                        type="bioimageio.core",
                        traceback=tb,
                    )
                ]
            ),
        )
    )


def _test_model_inference_parametrized(
    model: v0_5.ModelDescr,
    weight_format: WeightsFormat,
    devices: Optional[Sequence[str]],
) -> None:
    if not any(
        isinstance(a.size, v0_5.ParameterizedSize)
        for ipt in model.inputs
        for a in ipt.axes
    ):
        # no parameterized sizes => set n=0
        ns: Set[v0_5.ParameterizedSize_N] = {0}
    else:
        ns = {0, 1, 2}

    given_batch_sizes = {
        a.size
        for ipt in model.inputs
        for a in ipt.axes
        if isinstance(a, v0_5.BatchAxis)
    }
    if given_batch_sizes:
        batch_sizes = {gbs for gbs in given_batch_sizes if gbs is not None}
        if not batch_sizes:
            # only arbitrary batch sizes
            batch_sizes = {1, 2}
    else:
        # no batch axis
        batch_sizes = {1}

    test_cases: Set[Tuple[v0_5.ParameterizedSize_N, BatchSize]] = {
        (n, b) for n, b in product(sorted(ns), sorted(batch_sizes))
    }
    logger.info(
        "Testing inference with {} different input tensor sizes", len(test_cases)
    )

    def generate_test_cases():
        tested: Set[Hashable] = set()

        def get_ns(n: int):
            return {
                (t.id, a.id): n
                for t in model.inputs
                for a in t.axes
                if isinstance(a.size, v0_5.ParameterizedSize)
            }

        for n, batch_size in sorted(test_cases):
            input_target_sizes, expected_output_sizes = model.get_axis_sizes(
                get_ns(n), batch_size=batch_size
            )
            hashable_target_size = tuple(
                (k, input_target_sizes[k]) for k in sorted(input_target_sizes)
            )
            if hashable_target_size in tested:
                continue
            else:
                tested.add(hashable_target_size)

            resized_test_inputs = Sample(
                members={
                    t.id: test_inputs.members[t.id].resize_to(
                        {
                            aid: s
                            for (tid, aid), s in input_target_sizes.items()
                            if tid == t.id
                        },
                    )
                    for t in model.inputs
                },
                stat=test_inputs.stat,
                id=test_inputs.id,
            )
            expected_output_shapes = {
                t.id: {
                    aid: s
                    for (tid, aid), s in expected_output_sizes.items()
                    if tid == t.id
                }
                for t in model.outputs
            }
            yield n, batch_size, resized_test_inputs, expected_output_shapes

    try:
        test_inputs = get_test_inputs(model)

        with create_prediction_pipeline(
            bioimageio_model=model, devices=devices, weight_format=weight_format
        ) as prediction_pipeline:
            for n, batch_size, inputs, exptected_output_shape in generate_test_cases():
                error: Optional[str] = None
                result = prediction_pipeline.predict_sample_without_blocking(inputs)
                if len(result.members) != len(exptected_output_shape):
                    error = (
                        f"Expected {len(exptected_output_shape)} outputs,"
                        + f" but got {len(result.members)}"
                    )

                else:
                    for m, exp in exptected_output_shape.items():
                        res = result.members.get(m)
                        if res is None:
                            error = "Output tensors may not be None for test case"
                            break

                        diff: Dict[AxisId, int] = {}
                        for a, s in res.sizes.items():
                            if isinstance((e_aid := exp[AxisId(a)]), int):
                                if s != e_aid:
                                    diff[AxisId(a)] = s
                            elif (
                                s < e_aid.min or e_aid.max is not None and s > e_aid.max
                            ):
                                diff[AxisId(a)] = s
                        if diff:
                            error = (
                                f"(n={n}) Expected output shape {exp},"
                                + f" but got {res.sizes} (diff: {diff})"
                            )
                            break

                model.validation_summary.add_detail(
                    ValidationDetail(
                        name=f"Run {weight_format} inference for inputs with"
                        + f" batch_size: {batch_size} and size parameter n: {n}",
                        status="passed" if error is None else "failed",
                        errors=(
                            []
                            if error is None
                            else [
                                ErrorEntry(
                                    loc=("weights", weight_format),
                                    msg=error,
                                    type="bioimageio.core",
                                )
                            ]
                        ),
                    )
                )
    except Exception as e:
        error = str(e)
        tb = traceback.format_tb(e.__traceback__)
        model.validation_summary.add_detail(
            ValidationDetail(
                name=f"Run {weight_format} inference for parametrized inputs",
                status="failed",
                errors=[
                    ErrorEntry(
                        loc=("weights", weight_format),
                        msg=error,
                        type="bioimageio.core",
                        traceback=tb,
                    )
                ],
            )
        )


def _test_expected_resource_type(
    rd: Union[InvalidDescr, ResourceDescr], expected_type: str
):
    has_expected_type = rd.type == expected_type
    rd.validation_summary.details.append(
        ValidationDetail(
            name="Has expected resource type",
            status="passed" if has_expected_type else "failed",
            errors=(
                []
                if has_expected_type
                else [
                    ErrorEntry(
                        loc=("type",),
                        type="type",
                        msg=f"expected type {expected_type}, found {rd.type}",
                    )
                ]
            ),
        )
    )


def _handle_legacy_precision_args(
    absolute_tolerance: float, relative_tolerance: float, decimal: Optional[int]
) -> Dict[str, float]:
    """
    Transform the precision arguments to conform with the current implementation.

    If the deprecated `decimal` argument is used it overrides the new behaviour with
    the old behaviour.
    """
    # Already conforms with current implementation
    if decimal is None:
        return {
            "absolute_tolerance": absolute_tolerance,
            "relative_tolerance": relative_tolerance,
        }

    warnings.warn(
        "The argument `decimal` has been depricated in favour of "
        + "`relative_tolerance` and `absolute_tolerance`, with different validation "
        + "logic, using `numpy.testing.assert_allclose, see "
        + "'https://numpy.org/doc/stable/reference/generated/"
        + "numpy.testing.assert_allclose.html'. Passing a value for `decimal` will "
        + "cause validation to revert to the old behaviour."
    )
    # decimal overrides new behaviour,
    #   have to convert the params to emulate old behaviour
    return {
        "absolute_tolerance": 1.5 * 10 ** (-decimal),
        "relative_tolerance": 0,
    }


# def debug_model(
#     model_rdf: Union[RawResourceDescr, ResourceDescr, URI, Path, str],
#     *,
#     weight_format: Optional[WeightsFormat] = None,
#     devices: Optional[List[str]] = None,
# ):
#     """Run the model test and return dict with inputs, results, expected results and intermediates.

#     Returns dict with tensors "inputs", "inputs_processed", "outputs_raw", "outputs", "expected" and "diff".
#     """
#     inputs_raw: Optional = None
#     inputs_processed: Optional = None
#     outputs_raw: Optional = None
#     outputs: Optional = None
#     expected: Optional = None
#     diff: Optional = None

#     model = load_description(
#         model_rdf, weights_priority_order=None if weight_format is None else [weight_format]
#     )
#     if not isinstance(model, Model):
#         raise ValueError(f"Not a bioimageio.model: {model_rdf}")

#     prediction_pipeline = create_prediction_pipeline(
#         bioimageio_model=model, devices=devices, weight_format=weight_format
#     )
#     inputs = [
#         xr.DataArray(load_array(str(in_path)), dims=input_spec.axes)
#         for in_path, input_spec in zip(model.test_inputs, model.inputs)
#     ]
#     input_dict = {input_spec.name: input for input_spec, input in zip(model.inputs, inputs)}

#     # keep track of the non-processed inputs
#     inputs_raw = [deepcopy(input) for input in inputs]

#     computed_measures = {}

#     prediction_pipeline.apply_preprocessing(input_dict, computed_measures)
#     inputs_processed = list(input_dict.values())
#     outputs_raw = prediction_pipeline.predict(*inputs_processed)
#     output_dict = {output_spec.name: deepcopy(output) for output_spec, output in zip(model.outputs, outputs_raw)}
#     prediction_pipeline.apply_postprocessing(output_dict, computed_measures)
#     outputs = list(output_dict.values())

#     if isinstance(outputs, (np.ndarray, xr.DataArray)):
#         outputs = [outputs]

#     expected = [
#         xr.DataArray(load_array(str(out_path)), dims=output_spec.axes)
#         for out_path, output_spec in zip(model.test_outputs, model.outputs)
#     ]
#     if len(outputs) != len(expected):
#         error = f"Number of outputs and number of expected outputs disagree: {len(outputs)} != {len(expected)}"
#         print(error)
#     else:
#         diff = []
#         for res, exp in zip(outputs, expected):
#             diff.append(res - exp)

#     return {
#         "inputs": inputs_raw,
#         "inputs_processed": inputs_processed,
#         "outputs_raw": outputs_raw,
#         "outputs": outputs,
#         "expected": expected,
#         "diff": diff,
#     }
